from __future__ import annotations

"""
The AuthZero Configuration
"""

import os
import time
from pathlib import Path
from functools import lru_cache
from lazyops.utils.logs import logger as _logger, null_logger as _null_logger, Logger
from lazyops.imports._pydantic import BaseSettings
from lazyops.libs import lazyload
from lazyops.libs.proxyobj import ProxyObject

from lazyops.libs.abcs.configs.types import AppEnv
from lazyops.libs.fastapi_utils.types.persistence import TemporaryData
from pydantic import model_validator, computed_field, Field

from ..types.user_roles import UserRole
from ..utils.helpers import get_hashed_key, encrypt_key, decrypt_key, aencrypt_key, adecrypt_key, normalize_audience_name

from typing import List, Optional, Dict, Any, Union, overload, Callable, Tuple, TYPE_CHECKING

if lazyload.TYPE_CHECKING:
    import niquests
else:
    niquests = lazyload.LazyLoad("niquests")


lib_path = Path(__file__).parent.parent
_az_temp_data: Optional[TemporaryData] = None


@lru_cache()
def retrieve_jwks(domain: str) -> Dict[str, Any]:
    """
    Retrieves the JWKs from Auth0
    """
    url = f'https://{domain}/.well-known/jwks.json'
    attempts = 0
    e = None
    while attempts < 3:
        try:
            response = niquests.get(url)
            response.raise_for_status()
            return response.json()
        except Exception as e:
            attempts += 1
            _logger.warning(f"Unable to retrieve JWKS from {url}: {e}")
            time.sleep(3 * attempts)
    raise ValueError(f"Unable to retrieve JWKS from {url}")


def get_az_temp_data() -> TemporaryData:
    """
    Retrieves the temporary data
    """
    global _az_temp_data
    if _az_temp_data is None:
        _az_temp_data = TemporaryData(lib_path.joinpath('data', 'authzero.cache'))
    return _az_temp_data

class AuthZeroContextObject:
    """
    The Auth Zero Context
    """
    pre_validate_hooks: Optional[List[Callable]] = []
    post_validate_hooks: Optional[List[Callable]] = []
    configured_validators: List[str] = []

    validation_order: Optional[List[str]] = [
        'session',
        'api_key',
        'token'
    ]

    def add_post_validate_hook(self, hook: Callable):
        """
        Adds a post validate hook
        """
        if self.post_validate_hooks is None:
            self.post_validate_hooks = []
        self.post_validate_hooks.append(hook)
    
    def add_pre_validate_hook(self, hook: Callable):
        """
        Adds a pre validate hook
        """
        if self.pre_validate_hooks is None:
            self.pre_validate_hooks = []
        self.pre_validate_hooks.append(hook)

    def get_validation_hooks(self) -> Tuple[List[Callable], List[Callable]]:
        """
        Returns the validation hooks
        """
        return self.pre_validate_hooks.copy(), self.post_validate_hooks.copy()


AuthZeroContext: AuthZeroContextObject = ProxyObject(
    AuthZeroContextObject,
)

class AuthZeroSettings(BaseSettings):
    """
    The AuthZero Configuration
    """

    # Auth0 Client ID
    client_id: Optional[str] = None
    client_secret: Optional[str] = None
    audience: Optional[str] = None
    audiences: Optional[List[str]] = None

    apikey_header: Optional[str] = "x-api-key"
    authorization_header: Optional[str] = "authorization"
    authorization_scheme: Optional[str] = "bearer"

    domain: Optional[str] = None
    jwks_dict: Optional[Dict[str, Any]] = None

    admin_emails: Optional[List[str]] = None
    staff_email_domains: Optional[List[str]] = None

    # Persistence
    base_cache_key: Optional[str] = 'authzero'
    cache_key_prefix: Optional[str] = None
    data_dir: Optional[Path] = None
    local_persistence_fallback: Optional[bool] = True

    # API Keys
    api_key_prefix: Optional[str] = "xai-"
    api_key_secret_key: Optional[str] = None
    api_key_access_key: Optional[str] = None
    api_key_suffix: Optional[str] = None
    api_key_hashes: Optional[List[str]] = None

    # Can be a list or a dict of {key: role}
    if TYPE_CHECKING:
        allowed_api_keys: Optional[Union[List[str], Dict[str, Dict[str, Union[str, UserRole]]]]] = None
    else:
        allowed_api_keys: Optional[Union[str, List[str], Dict[str, str]]] = None


    # Metadata
    enabled: Optional[bool] = None
    session_enabled: Optional[bool] = None
    session_configure_enabled: Optional[bool] = None
    session_cookie_key: Optional[str] = None
    session_user_id_hashed: Optional[bool] = True
    enable_multi_audiences: Optional[bool] = None # allow for multiple audiences to be used

    # App Metadata - Used to identify the app
    app_name: Optional[str] = None
    app_env: Optional[Union[str, AppEnv]] = None
    app_ingress: Optional[str] = None
    app_client_id: Optional[str] = None

    app_scopes: Optional[List[str]] = None

    # API Clients
    api_client_id_header_key: Optional[str] = 'x-az-client-id'
    api_client_env_header_key: Optional[str] = 'x-az-client-env'
    api_client_api_key_prefix: Optional[str] = 'xaic-'

    # Expirations
    user_data_expiration: Optional[int] = 60 * 60 * 24 * 7 # 7 days
    user_session_expiration: Optional[int] = 60 * 60 * 24 * 30 # 30 days
    
    # Hook Configurations
    user_session_enabled: Optional[bool] = True
    user_request_id_enabled: Optional[bool] = True
    user_domain_source_enabled: Optional[bool] = False
    user_role_configure_enabled: Optional[bool] = True


    debug_enabled: Optional[bool] = False
    has_completed_validation: Optional[bool] = Field(None, exclude = True)
    
    if TYPE_CHECKING:
        ctx: Optional[AuthZeroContextObject] = None
    else:
        ctx: Optional[Any] = AuthZeroContext

    class Config:
        env_prefix = "AUTH_ZERO_"
        case_sensitive = False

    @computed_field
    @property
    def jwks(self) -> Dict[str, Any]:
        """
        Returns the JWKS Dict
        """
        if self.jwks_dict is not None: return self.jwks_dict
        if self.domain is None: raise ValueError("AuthZero Domain must be set")
        return retrieve_jwks(self.domain)

    @property
    def is_enabled(self) -> bool:
        """
        Returns whether the AuthZero is enabled
        """
        return self.enabled and self.client_id is not None and self.client_secret is not None


    @property
    def app_domain(self) -> Optional[str]:
        """
        Returns the App Domain
        """
        return self.app_ingress.replace('http://', '').replace('https://', '').rstrip('/') if self.app_ingress else None

    @property
    def is_secure_ingress(self) -> bool:
        """
        Returns whether the ingress is secure
        """
        return self.app_ingress.startswith("https://")

    @property
    def tenant(self) -> str:
        """
        Returns the Tenant
        """
        return f'https://{self.domain}'
    
    @property
    def issuer(self) -> str:
        """
        Returns the Issuer
        """
        return self.tenant.rstrip('/') + '/'
    
    @property
    def oauth_url(self) -> str:
        """
        Returns the OAuth URL
        """
        return f'{self.tenant}/oauth/token'
    
    @property
    def authorize_url(self) -> str:
        """
        Returns the Authorize URL
        """
        return f'{self.tenant}/authorize'
    
    @property
    def management_api_url(self) -> str:
        """
        Returns the Management API URL
        """
        return f'{self.tenant}/api/v2/'
    
    @property
    def api_key_enabled(self) -> bool:
        """
        Returns True if API Keys are Enabled
        """
        return bool(self.api_key_secret_key and self.api_key_access_key)
    

    def get_cache_key_prefix(self) -> str:
        """
        Returns the Cache Key Prefix
        """
        if self.cache_key_prefix: return self.cache_key_prefix
        if self.app_ingress: return get_hashed_key(self.app_domain)[:8]
        if self.app_name and self.app_env: return get_hashed_key(f'{self.app_name}-{self.app_env.name}')[:8]
        raise ValueError("Unable to determine cache key prefix, please set `app_name` and `app_env` or `app_ingress`")
    
    def get_cache_key(self, audience: str) -> str:
        """
        Returns the Cache Key
        """
        return f'{self.get_cache_key_prefix()}.{self.client_id[-10:]}.{normalize_audience_name(audience)}'


    def create_cache_key(
        self, 
        prefix: Optional[str] = None, 
        suffix: Optional[str] = None, 
        include_client_id: Optional[bool] = True,
        kind: Optional[str] = None
    ) -> str:
        """
        Creates a Cache Key based on the Prefix, Suffix, and Kind
        """
        # assert prefix is not None or kind is not None, "Either prefix or kind must be specified"
        cache_key_prefix = None
        if prefix is not None: cache_key_prefix = prefix
        elif kind is not None: cache_key_prefix = f'{self.get_cache_key_prefix()}.{kind}'
        else: cache_key_prefix = self.get_cache_key_prefix()
        # else: cache_key_prefix = f'{self.get_cache_key_prefix()}.{kind}'
        if include_client_id: cache_key_prefix = f'{cache_key_prefix}.{self.client_id[-10:]}'
        if suffix is not None: cache_key_prefix = f'{cache_key_prefix}.{suffix}'
        return cache_key_prefix

    def create_api_key(
        self,
        user_id: str,
        prefix: Optional[str] = None,
        suffix: Optional[str] = None,
    ) -> Optional[str]:
        """
        Creates an API Key
        """
        if not self.api_key_enabled: 
            raise ValueError("API Keys are not enabled. Please set `api_key_secret_key` and `api_key_access_key`")
        key = encrypt_key(user_id, self.api_key_secret_key, self.api_key_access_key)
        if prefix is None: prefix = self.api_key_prefix
        if prefix: key = f'{prefix}{key}'
        if suffix is None: suffix = self.api_key_suffix
        if suffix: key = f'{key}{suffix}'
        return key
    

    def decrypt_api_key(
        self,
        api_key: str,
        prefix: Optional[str] = None,
        suffix: Optional[str] = None,
    ) -> Optional[str]:
        """
        Decrypts an API Key
        """
        if not self.api_key_enabled: 
            raise ValueError("API Keys are not enabled. Please set `api_key_secret_key` and `api_key_access_key`")
        if prefix is None: prefix = self.api_key_prefix
        if prefix: api_key = api_key[len(prefix):]
        if suffix is None: suffix = self.api_key_suffix
        if suffix: api_key = api_key[:-len(suffix)]
        return decrypt_key(api_key, self.api_key_secret_key, self.api_key_access_key)
    

    """
    Async Methods
    """

    async def acreate_api_key(
        self,
        user_id: str,
        prefix: Optional[str] = None,
        suffix: Optional[str] = None,
    ) -> Optional[str]:
        """
        Creates an API Key
        """
        if not self.api_key_enabled: 
            raise ValueError("API Keys are not enabled. Please set `api_key_secret_key` and `api_key_access_key`")
        key = await aencrypt_key(user_id, self.api_key_secret_key, self.api_key_access_key)
        if prefix is None: prefix = self.api_key_prefix
        if prefix: key = f'{prefix}{key}'
        if suffix is None: suffix = self.api_key_suffix
        if suffix: key = f'{key}{suffix}'
        return key
    
    async def adecrypt_api_key(
        self,
        api_key: str,
        prefix: Optional[str] = None,
        suffix: Optional[str] = None,
    ) -> Optional[str]:
        """
        Decrypts an API Key
        """
        if not self.api_key_enabled: 
            raise ValueError("API Keys are not enabled. Please set `api_key_secret_key` and `api_key_access_key`")
        if prefix is None: prefix = self.api_key_prefix
        if prefix: api_key = api_key[len(prefix):]
        if suffix is None: suffix = self.api_key_suffix
        if suffix: api_key = api_key[:-len(suffix)]
        return await adecrypt_key(api_key, self.api_key_secret_key, self.api_key_access_key)


    def get_jwt_audiences(self) -> List[str]:
        """
        Returns the JWT Audiences
        """
        if self.enable_multi_audiences and self.audiences:
            return self.audiences
        return [self.audience]

    def add_allowed_api_key(self, key: str, role: Optional[UserRole] = None):
        """
        Adds an allowed api key
        """
        if self.allowed_api_keys is None: self.allowed_api_keys = {}
        self.allowed_api_keys[key] = UserRole.USER if role is None else UserRole.parse_role(role)

    """
    Misc Properties
    """

    @computed_field
    @property
    def in_k8s(self) -> bool:
        """
        Returns whether the app is running in kubernetes
        """
        from lazyops.utils.system import is_in_kubernetes
        return is_in_kubernetes()


    @property
    def is_local_env(self) -> bool:
        """
        Returns whether the environment is development
        """
        return self.app_env in [AppEnv.DEVELOPMENT, AppEnv.LOCAL] and not self.in_k8s
    
    @property
    def is_production_env(self) -> bool:
        """
        Returns whether the environment is production
        """
        return self.app_env == AppEnv.PRODUCTION and self.in_k8s

    @property
    def is_development_env(self) -> bool:
        """
        Returns whether the environment is development
        """
        return self.app_env in [AppEnv.DEVELOPMENT, AppEnv.LOCAL, AppEnv.CICD]
    
    @property
    def logger(self) -> 'Logger':
        """
        Returns the logger
        """
        return _logger
    
    @property
    def null_logger(self) -> 'Logger':
        """
        Returns the null logger
        """
        return _null_logger
    

    @property
    def autologger(self) -> 'Logger':
        """
        Returns the autologger
        """
        return self.logger if (self.is_development_env or self.debug_enabled) else self.null_logger


    
    """
    Validate/Configure the AuthZero Settings
    """

    def validate_audiences(self):
        """
        Validates the audience
        """
        if self.audiences:
            if self.enable_multi_audiences is None: self.enable_multi_audiences = True
            elif not self.enable_multi_audiences:
                raise ValueError("Multiple Audiences must be enabled to use multiple audiences")
            if self.audience and self.audience not in self.audiences:
                self.audiences.insert(0, self.audience)
        elif not self.audience:
            self.audience = f'https://{self.domain}/userinfo'

    def validate_app_ingress(self):
        """
        Validates the app ingress
        """
        if self.app_ingress is None: return
        if not self.app_ingress.startswith('http'):
            if 'localhost' in self.app_ingress or '127.0.0.1' in self.app_ingress or '0.0.0.0' in self.app_ingress:
                if not get_az_temp_data().has_logged('app_ingress_validate'):
                    self.logger.warning('`app_ingress` is not using https. This is insecure and is not recommended')
                self.app_ingress = f'http://{self.app_ingress}'
            else:
                self.app_ingress = f'https://{self.app_ingress}'
        self.app_ingress = self.app_ingress.rstrip('/')

    def parse_allowed_api_key(self, api_key: str) -> Tuple[str, str, UserRole]:
        """
        Parses the allowed api key

        Formats: 
        - {api_key}:{client_name}:{role}
        - {api_key}:{client_name}
        - {api_key}
        """
        parts = api_key.split(':')
        if len(parts) == 3:
            return parts[0], parts[1], UserRole.parse_role(parts[2])
        elif len(parts) == 2:
            return parts[0], parts[1], UserRole.API_CLIENT
        else:
            return api_key, 'default', UserRole.API_CLIENT


    def validate_allowed_api_keys(self):
        """
        Validates the allow api keys
        """
        if self.allowed_api_keys is None: return
        if isinstance(self.allowed_api_keys, str): self.allowed_api_keys = [self.allowed_api_keys]
        # Transform it to a dict
        if isinstance(self.allowed_api_keys, list):
            allowed_api_keys = {}
            for allowed_api_key in self.allowed_api_keys:
                key, client_name, role = self.parse_allowed_api_key(allowed_api_key)
                allowed_api_keys[key] = {
                    'client_name': client_name,
                    'role': role,
                }
            # allowed = {key: UserRole.USER for key in self.allowed_api_keys}
            self.allowed_api_keys = allowed_api_keys
        elif isinstance(self.allowed_api_keys, dict):
            # allowed_api_keys = {}
            for key, value in self.allowed_api_keys.items():
                if isinstance(value, str):  
                    self.allowed_api_keys[key] = {
                        'client_name': 'default',
                        'role': UserRole.parse_role(value),
                    }
                elif isinstance(value, dict) and 'role' not in value and 'client_name' not in value:
                    self.allowed_api_keys[key] = {
                        'client_name': value.get('client_name', 'default'),
                        'role': UserRole.parse_role(value.get('role', 'API_CLIENT')),
                    }
                
                # allowed_api_keys[key] = value
            # for key in self.allowed_api_keys:
            #     if self.allowed_api_keys[key] is None:
            #         self.allowed_api_keys[key] = UserRole.USER
            #     elif self.allowed_api_keys[key] not in UserRole:
            #         raise ValueError(f'Invalid User Role: {self.allowed_api_keys[key]}')

    def validate_app_scopes(self):
        """
        Validates the app scopes
        """
        if self.app_scopes is None: 
            self.app_scopes = ['openid', 'profile', 'email']
            return
        if not isinstance(self.app_scopes, list): self.app_scopes = [self.app_scopes]
        else: self.app_scopes = list(set(self.app_scopes))

    def validate_config(self):
        """
        Handles the validation of the configuration after updates / changes
        """
        if self.has_completed_validation: return
        if self.client_id is None: raise ValueError("AuthZero Client ID must be set")
        if self.client_secret is None: raise ValueError("AuthZero Client Secret must be set")
        if self.domain is None: raise ValueError("AuthZero Domain must be set")

        # Check API Keys
        # Validate that the API Key Secret Key is 16 characters
        tmp_data = get_az_temp_data()

        if not self.api_key_secret_key:
            if self.is_production_env: raise ValueError("API Key Secret Key must be set in production")
            if not tmp_data.has_logged('api_key_secret_key'):
                self.logger.warning("Using default `AUTH_ZERO_API_SECRET_KEY` [INSECURE]")
                self.api_key_secret_key = ('abc123' * 10)[:16]
        
        elif len(self.api_key_secret_key) != 16: raise ValueError("API Key Secret Key must be 16 characters")
        
        # Validate that the API Key Access Key is 16 characters
        if not self.api_key_access_key:
            if self.is_production_env: raise ValueError("API Key Access Key must be set in production")
            if not tmp_data.has_logged('api_key_access_key'):
                self.logger.warning("Using default `AUTH_ZERO_API_KEY_ACCESS_KEY` [INSECURE]")
                self.api_key_access_key = ('321bca' * 10)[:16]
        
        elif len(self.api_key_access_key) != 16: raise ValueError("API Key Access Key must be 16 characters")
        self.validate_audiences()
        self.validate_app_ingress()
        self.validate_allowed_api_keys()
        self.validate_app_scopes()
        # Validate Session Cookie Key
        if self.session_cookie_key is None:
            if self.app_name and self.app_env: self.session_cookie_key = f'{self.app_name}-{self.app_env.name}-session'.replace(' ', '-').lower()
            elif self.app_ingress: self.session_cookie_key = f'{self.app_ingress}-session'.split('://', 1)[-1].replace(' ', '-').replace('.', '').lower()

        self.has_completed_validation = True


    @model_validator(mode = 'after')
    def validate_authzero_config(self):
        """
        Validates the AuthZero Configuration
        """
        if self.data_dir is None: self.data_dir = lib_path.joinpath('data')
        if self.app_env is None: self.app_env = AppEnv.from_module_name('authzero')
        if self.ctx is None: self.ctx = AuthZeroContext
        if not self.enabled or not self.is_enabled: return self
        self.validate_config()
        return self


    @overload
    def configure(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        audience: Optional[str] = None,
        audiences: Optional[List[str]] = None,

        apikey_header: Optional[str] = None,
        authorization_header: Optional[str] = None,
        authorization_scheme: Optional[str] = None,

        domain: Optional[str] = None,
        jwks_dict: Optional[Dict[str, Any]] = None,

        # Persistence
        cache_key_prefix: Optional[str] = None,
        data_dir: Optional[Path] = None,

        # API Keys
        api_key_prefix: Optional[str] = None,
        api_key_secret_key: Optional[str] = None,
        api_key_access_key: Optional[str] = None,
        api_key_suffix: Optional[str] = None,
        api_key_hashes: Optional[List[str]] = None,

        # Metadata
        enabled: Optional[bool] = None,
        session_enabled: Optional[bool] = None,
        enable_multi_audiences: Optional[bool] = None, # allow for multiple audiences to be used

        # API Clients
        api_client_id_header_key: Optional[str] = None,
        api_client_env_header_key: Optional[str] = None,
        api_client_api_key_prefix: Optional[str] = None,

        # App Metadata - Used to identify the app
        app_name: Optional[str] = None,
        app_env: Optional[Union[AppEnv, str]] = None,
        app_ingress: Optional[str] = None,
        **kwargs,
    ):
        """
        Configure the AuthZero Settings
        """
        ...


    def configure(
        self,
        **kwargs,
    ):
        """
        Configure the AuthZero Settings
        """
        for k,v in kwargs.items():
            if k == 'app_env':
                if isinstance(v, str): v = AppEnv(v)
            if hasattr(self, k):
                setattr(self, k, v)    
        self.has_completed_validation = False
        self.validate_config()
        return self
    


settings: AuthZeroSettings = ProxyObject(
    obj_getter = 'lazyops.libs.authzero.utils.lazy.get_az_settings',
)


import copy
import threading
import contextlib
import operator
import copyreg
from typing import Any, Type, Tuple, Dict, List, Union, Optional, Callable, TypeVar, Iterable, Generic, TYPE_CHECKING


ProxyObjT = TypeVar('ProxyObjT')


class ProxyObjectV1(Generic[ProxyObjT]):
    def __init__(
        self,
        obj_cls: Optional[Union[Type[ProxyObjT], str]] = None,
        obj_getter: Optional[Union[Callable, str]] = None,
        obj_args: Optional[List[Any]] = None,
        obj_kwargs: Optional[Dict[str, Any]] = None,
        obj_initialize: Optional[bool] = True,
        threadsafe: Optional[bool] = True,
        debug_enabled: Optional[bool] = False,
    ):
        """
        Proxy Object

        args:
            obj_cls: the class of the object
            obj_getter: the function to get the object
            debug_enabled: if True, will raise an error if the object is not found
        """
        # Intentionally underscore on the end to avoid conflicts with the settings
        # Evaluate better approach with https://github.com/seperman/dotobject/blob/master/dot/borrowed_lazy.py
        assert obj_cls or obj_getter, "Either `obj_cls` or `obj_getter` must be provided"
        self.__obj_cls_ = obj_cls
        self.__obj_getter_ = obj_getter
        if self.__obj_getter_ and isinstance(self.__obj_getter_, str):
            from lazyops.utils.helpers import lazy_import
            self.__obj_getter_ = lazy_import(self.__obj_getter_)
        self.__threadlock_ = None if threadsafe else threading.Lock()
        self.__obj_ = None
        self.__obj_args_ = obj_args or []
        self.__obj_kwargs_ = obj_kwargs or {}
        self.__obj_initialize_ = obj_initialize
        self.__debug_enabled_ = debug_enabled
        self.__last_attrs_: Dict[str, int] = {}

    @contextlib.contextmanager
    def _objlock_(self):
        """
        Returns the object lock
        """
        if self.__threadlock_ is not None:
            try:
                with self.__threadlock_:
                    yield
            except Exception as e:
                raise e
        else:
            yield

    @property
    def _obj_(self) -> ProxyObjT:
        """
        Returns the object
        """
        if self.__obj_ is None:
            with self._objlock_():
                if self.__obj_getter_:
                    self.__obj_ = self.__obj_getter_(*self.__obj_args_, **self.__obj_kwargs_)
                elif self.__obj_cls_:
                    if isinstance(self.__obj_cls_, str):
                        from lazyops.utils.helpers import lazy_import
                        self.__obj_cls_ = lazy_import(self.__obj_cls_)
                    if self.__obj_initialize_:
                        self.__obj_ = self.__obj_cls_(*self.__obj_args_, **self.__obj_kwargs_)
                    else:
                        self.__obj_ = self.__obj_cls_
        return self.__obj_
    

    def __call__(self, *args, **kwargs) -> Any:
        """
        Call the proxy object
        """
        return self._obj_(*args, **kwargs)

    def __getattr__(self, name):
        """
        Forward all unknown attributes to the proxy object
        """
        if not self.__debug_enabled_:
            return getattr(self._obj_, name)
        
        # Try to debug the attribute
        if name not in self.__last_attrs_:
            self.__last_attrs_[name] = 0
        self.__last_attrs_[name] += 1
        if self.__last_attrs_[name] > 5:
            raise AttributeError(f"Proxy object has no attribute {name}")

        if hasattr(self._obj_, name):
            self.__last_attrs_[name] = 0
            return getattr(self._obj_, name)
        raise AttributeError(f"Settings object has no attribute {name}")
    
    if TYPE_CHECKING:
        def __new__(cls, *args, **kwargs) -> ProxyObjT:
            ...


class ProxyObjectV2(Generic[ProxyObjT]):
    def __init__(
        self,
        obj_cls: Optional[Union[Type[ProxyObjT], str]] = None,
        obj_getter: Optional[Union[Callable, str]] = None,
        obj_args: Optional[List[Any]] = None,
        obj_kwargs: Optional[Dict[str, Any]] = None,
        obj_initialize: Optional[bool] = True,
        threadsafe: Optional[bool] = True,
        debug_enabled: Optional[bool] = False,
    ) -> ProxyObjT:
        """
        Proxy Object

        args:
            obj_cls: the class of the object
            obj_getter: the function to get the object
            debug_enabled: if True, will raise an error if the object is not found
        """
        # Intentionally underscore on the end to avoid conflicts with the settings
        # Evaluate better approach with https://github.com/seperman/dotobject/blob/master/dot/borrowed_lazy.py
        assert obj_cls or obj_getter, "Either `obj_cls` or `obj_getter` must be provided"
        self.__dict__['__obj_cls_'] = obj_cls
        if obj_getter and isinstance(obj_getter, str):
            from lazyops.utils.helpers import lazy_import
            obj_getter = lazy_import(obj_getter)
        self.__dict__['__obj_getter_'] = obj_getter
        self.__dict__['__threadlock_'] = None if threadsafe else threading.Lock()
        self.__dict__['__obj_'] = None
        self.__dict__['__obj_args_'] = obj_args or []
        self.__dict__['__obj_kwargs_'] = obj_kwargs or {}
        self.__dict__['__obj_initialize_'] = obj_initialize
        self.__dict__['__debug_enabled_'] = debug_enabled
        self.__dict__['__last_attrs_'] = {}

        
    @contextlib.contextmanager
    def _objlock_(self):
        """
        Returns the object lock
        """
        if self.__dict__['__threadlock_'] is not None:
            try:
                with self.__dict__['__threadlock_']:
                    yield
            except Exception as e:
                raise e
        else:
            yield


    @property
    def _obj_(self) -> ProxyObjT:
        """
        Returns the object
        """
        if self.__dict__['__obj_'] is None:
            with self._objlock_():
                if self.__dict__['__obj_getter_']:
                    self.__dict__['__obj_'] = self.__dict__['__obj_getter_'](*self.__dict__['__obj_args_'], **self.__dict__['__obj_kwargs_'])
                elif self.__dict__['__obj_cls_']:
                    if isinstance(self.__dict__['__obj_cls_'], str):
                        from lazyops.utils.helpers import lazy_import
                        self.__dict__['__obj_cls_'] = lazy_import(self.__dict__['__obj_cls_'])
                    if self.__dict__['__obj_initialize_']:
                        self.__dict__['__obj_'] = self.__dict__['__obj_cls_'](*self.__dict__['__obj_args_'], **self.__dict__['__obj_kwargs_'])
                    else:
                        self.__dict__['__obj_'] = self.__dict__['__obj_cls_']
        return self.__dict__['__obj_']
    
    def __call__(self, *args, **kwargs) -> Any:
        """
        Call the proxy object
        """
        return self._obj_(*args, **kwargs)

    def __getattr__(self, name) -> Any:
        """
        Forward all unknown attributes to the proxy object
        """
        if name in self.__dict__:
            return self.__dict__[name]
        
        if not self.__dict__['__debug_enabled_']:
            return getattr(self._obj_, name)
        
        # Try to debug the attribute
        if name not in self.__dict__['__last_attrs_']:
            self.__dict__['__last_attrs_'][name] = 0
        self.__dict__['__last_attrs_'][name] += 1
        if self.__dict__['__last_attrs_'][name] > 5:
            raise AttributeError(f"Proxy object has no attribute {name}")
        if hasattr(self._obj_, name):
            self.__dict__['__last_attrs_'][name] = 0
            return getattr(self._obj_, name)
        raise AttributeError(f"Proxy object has no attribute {name}")
    
    if TYPE_CHECKING:
        def __new__(cls, *args, **kwargs) -> Type[ProxyObjT]:
            ...



class Constant(tuple):
    "Pretty display of immutable constant."

    def __new__(cls, name):
        return tuple.__new__(cls, (name,))

    def __repr__(self):
        return f'{self[0]}'

EMPTY = Constant('EMPTY')
empty = object()


def new_method_proxy(func):
    def inner(self: 'ProxyObjectV3', *args):
        if self._wrapped is empty:
            self._setup()
        return func(self._wrapped, *args)
    return inner

"""
Borrowed from 
https://github.com/seperman/dotobject/blob/master/dot/borrowed_lazy.py
"""


class ProxyObjectV3(Generic[ProxyObjT]):
    
    _wrapped = None

    if TYPE_CHECKING:
        def __new__(cls: Type[ProxyObjT], *args, **kwargs) -> ProxyObjT:
            ...

    def __init__(
        self,
        obj_cls: Optional[Union[Type[ProxyObjT], str]] = None,
        obj_getter: Optional[Union[Callable[..., ProxyObjT], str]] = None,
        obj_args: Optional[Union[str, Callable[..., Iterable[Any]], Iterable[Any]]] = None,
        obj_kwargs: Optional[Union[str, Callable[..., Dict[str, Any]], Dict[str, Any]]] = None,
        obj_initialize: Optional[bool] = True,
        threadsafe: Optional[bool] = True,
        # debug_enabled: Optional[bool] = False,
    ) -> ProxyObjT:
        """
        args:
            obj_cls: the class of the object
            obj_getter: the function to get the object
            debug_enabled: if True, will raise an error if the object is not found
        """
        
        assert obj_cls or obj_getter, "Either `obj_cls` or `obj_getter` must be provided"
        self._wrapped = empty
        self.__dict__['__obj_cls_'] = obj_cls
        # Defer until called.
        # if obj_getter and isinstance(obj_getter, str):
        #     from lazyops.utils.helpers import lazy_import
        #     obj_getter = lazy_import(obj_getter)
        # if obj_args and isinstance(obj_args, str):
        #     from lazyops.utils.helpers import lazy_import
        #     obj_args = lazy_import(obj_args)
        # if obj_kwargs and isinstance(obj_kwargs, str):
        #     from lazyops.utils.helpers import lazy_import
        #     obj_kwargs = lazy_import(obj_kwargs)
        self.__dict__['__obj_getter_'] = obj_getter
        self.__dict__['__threadlock_'] = None if threadsafe else threading.Lock()
        self.__dict__['__obj_args_'] = obj_args or []
        self.__dict__['__obj_kwargs_'] = obj_kwargs or {}
        self.__dict__['__obj_initialize_'] = obj_initialize
        # self.__dict__['__debug_enabled_'] = debug_enabled
        self.__dict__['__last_attrs_'] = {}

    
    @contextlib.contextmanager
    def _objlock_(self):
        """
        Returns the object lock
        """
        if self.__dict__['__threadlock_'] is not None:
            try:
                with self.__dict__['__threadlock_']:
                    yield
            except Exception as e:
                raise e
        else:
            yield

    __getattr__ = new_method_proxy(getattr)

    def __setattr__(self, name, value):
        if name == "_wrapped":
            # Assign to __dict__ to avoid infinite __setattr__ loops.
            self.__dict__["_wrapped"] = value
        else:
            if self._wrapped is empty:
                self._setup()
            setattr(self._wrapped, name, value)

    def __delattr__(self, name):
        if name == "_wrapped":
            raise TypeError("can't delete _wrapped.")
        if self._wrapped is empty:
            self._setup()
        delattr(self._wrapped, name)

    def __call__(self, *args, **kwargs) -> Any:
        """
        Call the proxy object
        """
        if self._wrapped is empty:
            self._setup()
        return self._wrapped(*args, **kwargs)

    def _setup_init(self):
        """
        Setup and initialize the proxy object arguments
        """
        from lazyops.utils.helpers import lazy_import
        if self.__dict__['__obj_args_'] is not None and not isinstance(self.__dict__['__obj_args_'], (list, tuple)):
            if isinstance(self.__dict__['__obj_args_'], str):
                self.__dict__['__obj_args_'] = lazy_import(self.__dict__['__obj_args_'])
            if callable(self.__dict__['__obj_args_']):
                self.__dict__['__obj_args_'] = self.__dict__['__obj_args_']()
        
        if self.__dict__['__obj_kwargs_'] is not None and not isinstance(self.__dict__['__obj_kwargs_'], dict):
            if isinstance(self.__dict__['__obj_kwargs_'], str):
                self.__dict__['__obj_kwargs_'] = lazy_import(self.__dict__['__obj_kwargs_'])
            if callable(self.__dict__['__obj_kwargs_']):
                self.__dict__['__obj_kwargs_'] = self.__dict__['__obj_kwargs_']()
        
        if self.__dict__['__obj_getter_'] is not None and isinstance(self.__dict__['__obj_getter_'], str):
            self.__dict__['__obj_getter_'] = lazy_import(self.__dict__['__obj_getter_'])

        elif self.__dict__['__obj_cls_'] is not None and isinstance(self.__dict__['__obj_cls_'], str):
            self.__dict__['__obj_cls_'] = lazy_import(self.__dict__['__obj_cls_'])


    def _setup(self):
        """
        Initializes the Proxy Object
        """
        # if self.__dict__['__obj_'] is not None: return
        
        with self._objlock_():
            self._setup_init()    
            if self.__dict__['__obj_getter_'] is not None:
                self.__dict__['_wrapped'] = self.__dict__['__obj_getter_'](*self.__dict__['__obj_args_'], **self.__dict__['__obj_kwargs_'])
            
            elif self.__dict__['__obj_cls_']:
                if self.__dict__['__obj_initialize_']:
                    self.__dict__['_wrapped'] = self.__dict__['__obj_cls_'](*self.__dict__['__obj_args_'], **self.__dict__['__obj_kwargs_'])
                else:
                    self.__dict__['_wrapped'] = self.__dict__['__obj_cls_']
        

    # Because we have messed with __class__ below, we confuse pickle as to what
    # class we are pickling. It also appears to stop __reduce__ from being
    # called. So, we define __getstate__ in a way that cooperates with the way
    # that pickle interprets this class.  This fails when the wrapped class is
    # a builtin, but it is better than nothing.
    def __getstate__(self):
        if self._wrapped is empty:
            self._setup()
        return self._wrapped.__dict__

    # Python 3.3 will call __reduce__ when pickling; this method is needed
    # to serialize and deserialize correctly.
    @classmethod
    def __newobj__(cls, *args):
        return cls.__new__(cls, *args)
    
    def __reduce_ex__(self, proto):
        return (self.__newobj__, (self.__class__,), self.__getstate__())


    def __deepcopy__(self, memo):
        if self._wrapped is empty:
            # We have to use type(self), not self.__class__, because the
            # latter is proxied.
            result = type(self)()
            memo[id(self)] = result
            return result
        return copy.deepcopy(self._wrapped, memo)

    __bytes__ = new_method_proxy(bytes)
    __str__ = new_method_proxy(str)
    __bool__ = new_method_proxy(bool)
    # Introspection support
    __dir__ = new_method_proxy(dir)

    # Need to pretend to be the wrapped class, for the sake of objects that
    # care about this (especially in equality tests)
    __class__ = property(new_method_proxy(operator.attrgetter("__class__")))
    __eq__ = new_method_proxy(operator.eq)
    __ne__ = new_method_proxy(operator.ne)
    __hash__ = new_method_proxy(hash)

    # Dictionary methods support
    __getitem__ = new_method_proxy(operator.getitem)
    __setitem__ = new_method_proxy(operator.setitem)
    __delitem__ = new_method_proxy(operator.delitem)

    __len__ = new_method_proxy(len)
    __contains__ = new_method_proxy(operator.contains)

    # Additions for DotObject
    __gt__ = new_method_proxy(operator.gt)
    __lt__ = new_method_proxy(operator.lt)
    __ge__ = new_method_proxy(operator.ge)
    __le__ = new_method_proxy(operator.le)
    __add__ = new_method_proxy(operator.add)
    __radd__ = new_method_proxy(operator.add)
    __sub__ = new_method_proxy(operator.sub)
    __rsub__ = new_method_proxy(operator.sub)
    __mul__ = new_method_proxy(operator.mul)
    __rmul__ = new_method_proxy(operator.mul)
    __floordiv__ = new_method_proxy(operator.floordiv)
    __div__ = new_method_proxy(operator.truediv)
    __rdiv__ = new_method_proxy(operator.truediv)
    __truediv__ = new_method_proxy(operator.truediv)
    __rtruediv__ = new_method_proxy(operator.truediv)
    __mod__ = new_method_proxy(operator.mod)
    __rmod__ = new_method_proxy(operator.mod)
    __pow__ = new_method_proxy(operator.pow)
    __rpow__ = new_method_proxy(operator.pow)
    __lshift__ = new_method_proxy(operator.lshift)
    __rshift__ = new_method_proxy(operator.rshift)
    __and__ = new_method_proxy(operator.and_)
    __or__ = new_method_proxy(operator.or_)
    __xor__ = new_method_proxy(operator.xor)




ProxyObject = ProxyObjectV3


from __future__ import annotations

"""
Persistence Types
"""

from pydantic import BaseModel
from abc import ABC, abstractmethod
from typing import TypeVar, Generic, Any, Dict, Optional, Iterable, Type, TYPE_CHECKING

SchemaType = TypeVar("SchemaType", bound=BaseModel)

if TYPE_CHECKING:
    from kvdb import PersistentDict
    from lazyops.libs.authzero.configs import AuthZeroSettings


class StatefulProperty(Generic[SchemaType], ABC):
    """
    Router Class for Stateful Properties
    """
    __obj_: Optional[SchemaType] = None
    _settings: Optional['AuthZeroSettings'] = None
    name: Optional[str] = None
    schema_type: Type[BaseModel] = None

    def __init__(
        self, 
        cache_key: str,
        serializer: Optional[str] = "json",
        serializer_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        """
        Initializes the Stateful Property
        """
        self.cache_key = cache_key
        self._pdict: Optional['PersistentDict'] = None
        self._pdict_kwargs: Dict[str, Any] = {
            'serializer': serializer,
            'serializer_kwargs': serializer_kwargs,
            **kwargs,
        }
        self.data_cache_key = f'{self.cache_key}:data'

    @property
    def settings(self) -> 'AuthZeroSettings':
        """
        Returns the settings
        """
        if self._settings is None:
            from lazyops.libs.authzero.utils.lazy import get_az_settings
            self._settings = get_az_settings()
        return self._settings
    
    @property
    def pdict(self) -> 'PersistentDict':
        """
        Returns the Persistent Dict
        """
        if self._pdict is None:
            from lazyops.libs.authzero.utils.lazy import get_az_pdict
            from lazyops.libs.authzero.utils.helpers import normalize_audience_name
            base_key = f'{self.settings.base_cache_key}.'
            if self.settings.app_name:
                base_key += f'{self.settings.app_name}.'
            elif self.settings.app_ingress:
                base_key += f'{normalize_audience_name(self.settings.app_ingress)}.'
            else: base_key += 'default.'
            base_key += f'{self.settings.app_env.name}.{self.name}'
            base_key = base_key.replace(' ', '_').lower().replace('..', '.')
            self._pdict = get_az_pdict(
                base_key = base_key,
                **self._pdict_kwargs,
            )
        return self._pdict

    @property
    def resource(self) -> SchemaType:
        """
        Returns the Object Resource
        """
        # if self.__obj_ is None or (self.__obj_ is not None and self.is_expired(self.__obj_)):
        if self.__obj_ is None or self.is_expired(self.__obj_):
            self.__obj_ = self.fetch()
        return self.__obj_
    
    @property
    async def aresource(self) -> SchemaType:
        """
        Returns the Object Resource
        """
        if self.__obj_ is None or self.is_expired(self.__obj_):
            self.__obj_ = await self.afetch()
        return self.__obj_

    def __call__(self, *args, **kwargs) -> Any:
        """
        Call the proxy object
        """
        return self.resource(*args, **kwargs)

    def __getattr__(self, name):
        """
        Forward all unknown attributes to the proxy object
        """
        if hasattr(self.resource, name):
            return getattr(self.resource, name)
        raise AttributeError(f"Property {name} not found in {self.name}")

    def fetch(self) -> SchemaType:
        """
        Returns the Value
        """
        data = self.pdict.get(self.cache_key)
        if data is None or not self.is_valid_type(data) or self.is_expired(data):
            data = self.load()
            self.pdict.set(self.cache_key, data, ex = self.get_ttl(data))
        return data
    
    async def afetch(self) -> SchemaType:
        """
        Returns the Value
        """
        data = await self.pdict.aget(self.cache_key)
        if data is None or not self.is_valid_type(data) or self.is_expired(data):
            data = await self.aload()
            await self.pdict.aset(self.cache_key, data, ex = self.get_ttl(data))
        return data
    
    @classmethod
    def get(cls, *args, **kwargs) -> SchemaType:
        """
        Get the value
        """
        return cls(*args, **kwargs).fetch()
    
    @classmethod
    async def aget(cls, *args, **kwargs) -> SchemaType:
        """
        Get the value
        """
        return await cls(*args, **kwargs).afetch()

    def is_valid_type(self, data: Any) -> bool:
        """
        Returns True if the Data is a Valid Type
        """
        return self.schema_type is None or isinstance(data, self.schema_type)

    async def aload(self) -> SchemaType:
        """
        Loads the Data
        """
        raise NotImplementedError

    def load(self) -> SchemaType:
        """
        Loads the Data
        """
        raise NotImplementedError

    def is_expired(self, data: SchemaType) -> bool:
        """
        Returns True if the Data is Expired
        """
        raise getattr(data, 'is_expired', False)

    def get_ttl(self, data: SchemaType) -> int:
        """
        Returns the TTL
        """
        return None
    
    def save(self, data: SchemaType, ex: Optional[int] = None) -> None:
        """
        Saves the Object
        """
        self.pdict.set(self.cache_key, data, ex = ex or self.get_ttl(data))

    async def asave(self, data: SchemaType, ex: Optional[int] = None) -> None:
        """
        Saves the Object
        """
        await self.pdict.aset(self.cache_key, data, ex = ex or self.get_ttl(data))

    def delete(self):
        """
        Deletes the Object
        """
        self.pdict.delete(self.cache_key)

    async def adelete(self):
        """
        Deletes the Object
        """
        await self.pdict.adelete(self.cache_key)

    def save_data(self, data: Any, ex: Optional[int] = None) -> None:
        """
        Saves the Data
        """
        self.pdict.set(self.data_cache_key, data, ex = ex)

    async def asave_data(self, data: Any, ex: Optional[int] = None) -> None:
        """
        Saves the Data
        """
        await self.pdict.aset(self.data_cache_key, data, ex = ex)

    def load_data(self, default: Optional[Any] = None) -> Any:
        """
        Loads the Data
        """
        return self.pdict.get(self.data_cache_key, default = default)
    
    async def aload_data(self, default: Optional[Any] = None) -> Any:
        """
        Loads the Data
        """
        return await self.pdict.aget(self.data_cache_key, default = default)
    
    def delete_data(self) -> None:
        """
        Deletes the Data
        """
        self.pdict.delete(self.data_cache_key)
    
    async def adelete_data(self) -> None:
        """
        Deletes the Data
        """
        await self.pdict.adelete(self.data_cache_key)

    @classmethod
    def create_hash_key(cls, data: Any) -> str:
        """
        Creates a hash for the data
        """
        from ..utils.helpers import get_hashed_key
        return get_hashed_key(data)
    
    
    




from __future__ import annotations

"""
Lazy Loading Utilities
"""

from lazyops.libs import lazyload
from lazyops.libs.pooler import ThreadPooler
from lazyops.utils.logs import logger
from lazyops.utils.lazy import lazy_import
from lazyops.utils.helpers import fail_after
from typing import Any, Callable, Dict, List, Optional, Union, Type

if lazyload.TYPE_CHECKING:
    import kvdb
    from lazyops.libs.authzero.configs import AuthZeroSettings
    from kvdb import KVDBSession, PersistentDict

    from ..types import AZResource, AZResourceSchema
    from ..flows import AZFlow, AZFlowSchema, AZManagementAPI

else:
    kvdb = lazyload.LazyLoad("kvdb")
    # KVDBClient = lazyload.LazyLoad("kvdb.KVDBClient", package = 'kvdb')


_az_sessions_configured: Optional[bool] = False
_az_kdbs: Dict[str, 'KVDBSession'] = {}
_az_pdicts: Dict[str, 'PersistentDict'] = {}
_az_pdict_aliases: Dict[str, str] = {}
_az_settings: Optional['AuthZeroSettings'] = None
_az_mtg_api: Optional['AZManagementAPI'] = None

def get_az_kdb(
    name: Optional[str] = None,
    serializer: Optional[str] = 'json',
    **kwargs,
) -> 'KVDBSession':
    """
    Returns the KVDB Session
    """
    global _az_kdbs
    if name is None: name = 'global'
    if name not in _az_kdbs:
        _az_kdbs[name] = kvdb.KVDBClient.get_session(
            name = name,
            serializer = serializer,
            **kwargs,
        )
    return _az_kdbs[name]

def get_az_pdict(
    base_key: str,
    expiration: Optional[int] = None,
    aliases: Optional[List[str]] = None,
    hset_disabled: Optional[bool] = False,
    **kwargs,
) -> 'PersistentDict':
    """
    Lazily initializes the persistent dict
    """
    global _az_pdicts, _az_pdict_aliases
    if base_key not in _az_pdicts and base_key not in _az_pdict_aliases:
        s = get_az_settings()
        sess = get_az_kdb(
            'persistence',
            serializer = None,
        )
        if s.local_persistence_fallback:
            try:
                with fail_after(5):
                    sess.ping()
            except Exception as e:
                logger.warning(f'Failed to connect to KVDB persistence backend, falling back to local: {e}')
        if sess is not None:
            _az_pdicts[base_key] = sess.create_persistence(
                base_key = base_key,
                expiration = expiration,
                hset_disabled = hset_disabled,
                **kwargs,
            )
        else:
            from lazyops.libs.persistence import PersistentDict
            _az_pdicts[base_key] = PersistentDict(
                base_key = base_key,
                expiration = expiration,
                hset_disabled = hset_disabled,
                file_path = s.data_dir.joinpath(f'{s.app_name}.cache'),
                **kwargs,
            )

        if aliases:
            for alias in aliases:
                _az_pdict_aliases[alias] = base_key
    elif base_key in _az_pdict_aliases:
        base_key = _az_pdict_aliases[base_key]
    return _az_pdicts[base_key]


def get_az_settings() -> 'AuthZeroSettings':
    """
    Returns the AuthZero Settings
    """
    global _az_settings
    if _az_settings is None:
        from ..configs import AuthZeroSettings
        _az_settings = AuthZeroSettings()
    return _az_settings


_az_flow_schemas: Dict[str, 'AZFlowSchema'] = {
    'api_key': 'lazyops.libs.authzero.flows.api_keys.APIKeyDataFlow',
    'api_client_credentials': 'lazyops.libs.authzero.flows.tokens.APIClientCredentialsFlow',
    'client_credentials': 'lazyops.libs.authzero.flows.tokens.ClientCredentialsFlow',
    'user_data': 'lazyops.libs.authzero.flows.user_data.UserDataFlow',
    'user_session': 'lazyops.libs.authzero.flows.user_session.UserSessionFlow',
    
}

def get_az_flow_schema(
    name: str
) -> 'AZFlowSchema':
    """
    Returns the AZFlowSchema
    """
    global _az_flow_schemas
    if name not in _az_flow_schemas:
        raise ValueError(f"Invalid AuthZero Flow: {name}, must be one of {list(_az_flow_schemas.keys())}")
    if isinstance(_az_flow_schemas[name], str):
        _az_flow_schemas[name] = lazy_import(_az_flow_schemas[name])
    return _az_flow_schemas[name]

def get_az_flow(
    name: str, 
    *args, 
    **kwargs
) -> 'AZFlow':
    """
    Returns the AZFlow
    """
    return get_az_flow_schema(name)(*args, **kwargs)


_az_resource_schemas: Dict[str, 'AZResourceSchema'] = {
    'current_user': 'lazyops.libs.authzero.types.current_user.CurrentUser',
    'user_role': 'lazyops.libs.authzero.types.user_roles.UserRole',
    'user_data': 'lazyops.libs.authzero.types.user_data.AZUserData',
    'auth_object': 'lazyops.libs.authzero.types.auth.AuthObject',
    'user_jwt_claims': 'lazyops.libs.authzero.types.claims.UserJWTClaims',
    'api_key_jwt_claims': 'lazyops.libs.authzero.types.claims.APIKeyJWTClaims',
    'auth_zero_token_auth': 'lazyops.libs.authzero.types.auth.AuthZeroTokenAuth',
    'user_session': 'lazyops.libs.authzero.types.user_session.UserSession',
}

def get_az_resource_schema(
    name: str,
) -> 'AZResourceSchema':
    """
    Returns the AZResource Schema
    """
    global _az_resource_schemas
    if name not in _az_resource_schemas:
        raise ValueError(f"Invalid AuthZero Resource: {name}, must be one of {list(_az_resource_schemas.keys())}")
    if isinstance(_az_resource_schemas[name], str):
        _az_resource_schemas[name] = lazy_import(_az_resource_schemas[name])
    return _az_resource_schemas[name]


def get_az_resource(
    name: str,
    *args, 
    **kwargs
) -> 'AZResource':
    """
    Returns the AZResource
    """
    schema = get_az_resource_schema(name)
    return schema(*args, **kwargs)


def get_az_mtg_api() -> 'AZManagementAPI':
    """
    Returns the AZ Management API
    """
    global _az_mtg_api
    if _az_mtg_api is None:
        from ..flows.admin import AZManagementClient
        _az_mtg_api = AZManagementClient
    return _az_mtg_api


from __future__ import annotations

import contextlib
from lazyops.types import BaseModel, Field, root_validator
from lazyops.types.models import ConfigDict, schema_extra

from fastapi.responses import JSONResponse
from fastapi.exceptions import HTTPException
from kvdb.types.jobs import Job, JobStatus
from lazyops.libs.logging import logger
from typing import Any, Dict, List, Optional, Type, TypeVar, Literal, Union, Set, TYPE_CHECKING

if TYPE_CHECKING:
    from lazyops.libs.abcs.configs.base import AppSettings
    from lazyops.libs.abcs.types.errors import APIException

class BaseSchema(BaseModel):
    """
    BaseModel with some extra features, designed to be used
    for Pydantic Serializability in FastAPI
    """

    @classmethod
    def schema(cls, by_alias: bool = False, ref_template: str = None, **kwargs) -> Dict[str, Any]:
        """
        Returns the schema
        """
        if ref_template is None:
            ref_template = '#/components/schemas/{model}'
        return cls.model_json_schema(by_alias = by_alias, ref_template = ref_template, **kwargs)

    @property
    def settings(self) -> 'AppSettings':
        """
        Returns the app settings
        """
        from lazyops.libs.abcs.configs.lazy import get_module_settings
        return get_module_settings(self.__module__.__name__)
    
    @property
    def status_code(self) -> int:
        """
        Returns the status code
        """
        return 200
    
    @property
    def response_headers(self) -> Optional[Dict[str, Any]]:
        """
        Returns the response headers
        """
        return {}
    
    @property
    def schema_object_index_key(self) -> str:
        """
        Returns the schema object index key
        """
        return None
    
    @property
    def _excluded_fields(self) -> List[str]:
        """
        Returns the excluded fields
        """
        fields = []
        for field_name, field in self.model_fields.items():
            if field_name.startswith('_'): fields.append(field_name)
            elif getattr(field, 'is_excluded', None): fields.append(field_name)
        return fields
    
    def get_excluded_fields(self, exclude: Optional[Set[str]] = None) -> Set[str]:
        """
        Returns the excluded fields
        """
        exclude = exclude or set()
        exclude.update(self._excluded_fields)
        return exclude
    
    def json_response(self, exclude_none: bool = True, exclude: Optional[Set[str]] = None, indent: Optional[int] = 2, **kwargs) -> JSONResponse:
        """
        Returns the JSON response
        """
        return JSONResponse(
            content = self.dict(
                exclude = self.get_excluded_fields(exclude), 
                exclude_none = exclude_none, 
                **kwargs, 
            ),
            status_code = self.status_code,
            headers = self.response_headers,
        )

    def dict(self, exclude_none: bool = True, **kwargs):
        """
        We want to exclude None values by default
        """
        return self.model_dump(exclude_none = exclude_none, **kwargs)

    @classmethod
    def schema_name(cls) -> str:
        """
        Returns the class name
        """
        return f'{cls.__module__}.{cls.__name__}'
    
    
    @classmethod
    def raise_http_exception(cls, message: str, status_code: int = 400, **kwargs) -> None:
        """
        Raises an HTTPException
        """
        raise HTTPException(
            status_code = status_code,
            detail = message,
            **kwargs
        )
    

    def partial_log(
        self,
        limit: Optional[int] = None,
        fields: Optional[List[str]] = None,
        pretty: Optional[bool] = True,
        colored: Optional[bool] = False,
    ) -> str:
        """
        Returns the partial log as a string
        """
        s = f'[{self.__class__.__name__}]'
        if colored: s = f'|g|{s}|e|'
        if fields is None: fields = self.get_model_field_names()
        for field in fields:
            field_str = f'|g|{field}|e|' if colored else field
            val_s = f'\n\t{field_str}: {getattr(self, field)!r}' if pretty else f'{field_str}={getattr(self, field)!r}, '
            if limit is not None and len(val_s) > limit:
                val_s = f'{val_s[:limit]}...'
            s += val_s
        return s
    

    @contextlib.contextmanager
    def catch(
        self, 
        exc: Type['APIException'], 
        msg: Optional[str] = None, 
        fatal: Optional[bool] = False
    ):
        """
        Catches an exception
        """
        try:
            yield
        except Exception as e:
            logger.trace(f'Handling Error: {msg}', e)
            if fatal: raise exc(msg, error = e) from e

    def prepare_response(self, mode: str = 'python', **kwargs) -> Any:
        """
        Prepares the response
        """
        return self.model_dump(mode = mode, **kwargs)


    model_config = ConfigDict(json_schema_extra=schema_extra, extra='allow', arbitrary_types_allowed=True)



class BaseResponse(BaseSchema):
    """
    Response Schema for Requests
    """
    request_id: Optional[str] = Field(None, description = 'The request ID')
    callback_id: Optional[str] = Field(None, description = 'The callback ID')

    @property
    def response_headers(self) -> Optional[Dict[str, Any]]:
        """
        Returns the response headers
        """
        if not self.request_id and not self.callback_id: return None
        headers = {}
        if self.request_id: headers['X-Request-ID'] = self.request_id
        if self.callback_id: headers['X-Callback-ID'] = self.callback_id
        return headers


class JobResult(BaseResponse):

    status: JobStatus = Field(..., description = 'The job status')
    progress: Optional[float] = Field(None, description = 'The job progress')
    duration: Optional[float] = Field(None, description = 'The job duration in seconds')

    model_config = {'extra': 'ignore', 'arbitrary_types_allowed': True}

    
    @classmethod
    def from_job(
        cls, 
        job: Job, 
        callback_id: Optional[str] = None,
        **kwargs,
    ) -> 'JobResult':
        """
        Returns the JobResult from the Job
        """
        return cls(
            job_id = job.id,
            status = job.status,
            progress = job.progress.completed,
            duration = job.duration / 1000 if job.duration else None,
            callback_id = callback_id,
            **kwargs,
        )


SchemaT = TypeVar('SchemaT', bound = BaseSchema)
ResultT = Union[JobResult, SchemaT, Dict[str, Any], Any]


class BaseResponse(BaseResponse):
    """
    Base Response Schema for Tasks/Services
    """

    object: Optional[str] = Field(None, description = 'The Response Schema object')
    status: Optional[JobStatus] = Field(None, description = 'The Current Task Status')
    result: Optional[Union[Dict[str, Any], List[Union[Dict[str, Any], Any]], str]] = Field(None, description = 'The Result Object')
    error: Optional[Any] = Field(None, description = 'The Error Message or Exception')

    request_id: Optional[str] = Field(None, description = 'The request ID')
    callback_id: Optional[str] = Field(None, description = 'The callback ID')


    @root_validator(pre = True)
    def validate_object_schema(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """
        Validate that the schema is valid
        """
        # logger.info(f'Validating Response Schema: {values}')
        result = values.get('result')
        if result:
            if hasattr(result, 'is_exception') or isinstance(
                result, (Exception, HTTPException)
            ):
                values['object'] = result.__class__.__name__
                values['error'] = f'{result.__class__.__name__}: {str(result)}' 
                result = None

            elif hasattr(result, 'job_id'): result = JobResult(**result.dict())
            values['result'] = result

        if not values.get('object'):
            if result and hasattr(result, '__class__'): values['object'] = result.__class__.__name__
            elif values.get('error'): values['object'] = 'Error'

        if values.get('status') == JobStatus.FAILED:
            values['object'] = 'Error' 
            values['error'] = values.get('result')

        return values
    
    @property
    def response_headers(self) -> Optional[Dict[str, Any]]:
        """
        Returns the response headers
        """
        if not self.request_id and not self.callback_id: return None
        headers = {}
        if self.request_id: headers['X-Request-ID'] = self.request_id
        if self.callback_id: headers['X-Callback-ID'] = self.callback_id
        return headers


    @property
    def job_id(self):
        """
        Returns the job ID
        """
        return getattr(self.result, 'job_id', None) if self.result else None
    
    @property
    def is_error(self):
        """
        Returns True if the response has an error
        """
        return bool(self.error) and self.status == JobStatus.FAILED and \
            (self.result and isinstance(self.result, (Exception, HTTPException)))

    @property
    def has_error(self):
        """
        Returns True if the response has an error
        """
        return bool(self.error)
    
    @property
    def is_job(self):
        """
        Returns True if the response is a job
        """
        return bool(self.job_id)
    
    @property
    def is_result(self):
        """
        Returns True if the response is a result
        """
        return not self.is_job
    
    @property
    def is_complete(self):
        """
        Returns True if the response is complete
        """
        return self.status == JobStatus.COMPLETE and \
            self.is_result
    
    @property
    def status_code(self):
        """
        Returns the status code
        """
        if self.has_error or self.status == JobStatus.FAILED or self.object == "Error": return 400
        return 200 if self.status == JobStatus.COMPLETE else 202
    
    def raise_for_exceptions(self):
        """
        Helper method to raise for exceptions
        """
        if not self.has_error: return
        raise HTTPException(
            status_code = self.status_code,
            detail = f'{self.error}'
        )
    

    def __bool__(self):
        """
        Returns True if the response is successful
        """
        return not self.error and self.status in [
            JobStatus.COMPLETE,
            JobStatus.ACTIVE,
            JobStatus.QUEUED,
            JobStatus.NEW,
        ]


class BaseRequest(BaseSchema):
    """
    Base Request Schema 
    """

    def get_excluded_function_params(
        self,
        key: Optional[str] = 'exclude_param',
    ) -> List[str]:
        """
        Returns the excluded param fields
        """
        return [
            field_name
            for field_name, field in self.model_fields.items()
            if field.json_schema_extra.get(key, False)
        ]



    def get_function_kwargs(
        self,
        exclude_unset: Optional[bool] = True,
        exclude: Optional[Set[str]] = None,
        **kwargs
    ) -> Dict[str, Any]:
        """
        Returns the Function Kwargs
        """
        exclude = exclude or set()
        excluded_params = self.get_excluded_function_params()
        if excluded_params: exclude.update(excluded_params)
        return self.model_dump(
            exclude_unset = exclude_unset,
            exclude = exclude,
            **kwargs,
        )

    @classmethod
    def from_get_request(
        cls,
        **kwargs,
    ) -> 'BaseRequest':
        """
        Builds the BaseRequest from the GET Request
        """
        from lazyops.utils.helpers import build_dict_from_query
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        for key in kwargs:
            if not cls.model_fields.get(key): continue
            # Check if the value is a dict
            annotation = cls.model_fields[key].annotation
            if isinstance(kwargs[key], str) and \
                (
                    annotation.__origin__ == dict or \
                    (
                        annotation.__origin__ == Union and \
                        'typing.Dict' in str(annotation.__args__[0])
                    )
                ):
                kwargs[key] = build_dict_from_query(kwargs[key])

        return cls(**kwargs)
    


class CallbackRequestSchema(BaseRequest):
    """
    Parameters for Callbacks upon Completion
    """
    callback_url: Optional[str] = Field(None, description="If provided, will send a callback notification to this endpoint when the job is complete", exclude = True)
    callback_id: Optional[str] = Field(None, description="If provided, will use this callback id when sending a callback notification to the callback_url", exclude = True)
    callback_method: Optional[str] = Field('POST', description="If provided, will use this method when sending a callback notification to the callback_url", exclude = True)
    callback_params: Optional[Dict[str, Any]] = Field(None, description="If provided, will use these parameters when sending a callback notification to the callback_url", exclude = True)
    callback_headers: Optional[Dict[str, Any]] = Field(None, description="If provided, will use these headers when sending a callback notification to the callback_url", exclude = True)
    callback_retries: Optional[int] = Field(None, description="If provided, will retry the callback this many times before giving up", exclude = True)
    callback_timeout: Optional[int] = Field(None, description="If provided, will timeout the callback after this many seconds", exclude = True)
    callback_refire: Optional[bool] = Field(None, description="If provided, will refire the callback when the result is retrieved again", exclude = True)

    @property
    def callback_enabled(self) -> bool:
        """
        Returns True if callback_url is not None
        """
        return self.callback_url is not None
    
    @property
    def callback_param_fields(self) -> List[str]:
        """
        Returns the param fields
        """
        return [
            'callback_url',
            'callback_id',
            'callback_method',
            'callback_params',
            'callback_headers',
            'callback_retries',
            'callback_timeout',
            'callback_refire',
        ]
    
    @classmethod
    def from_request(
        cls, 
        **kwargs,
    ) -> 'CallbackRequestSchema':
        """
        Parses the keywords from the request
        """
        from lazyops.utils.helpers import build_dict_from_query
        if callback_params := kwargs.pop('callback_params', None):
            kwargs['callback_params'] = build_dict_from_query(callback_params)
        if callback_headers := kwargs.pop('callback_headers', None):
            kwargs['callback_headers'] = build_dict_from_query(callback_headers)
        return cls(**kwargs)
    
    def to_request_kwargs(
        self,
        exclude_none: bool = True,
        exclude_unset: bool = True,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Returns the request kwargs
        """
        return self.model_dump(
            include = set(self.callback_param_fields),
            exclude_none = exclude_none,
            exclude_unset = exclude_unset,
            **kwargs,
        )



from __future__ import annotations

"""
Create the App Context
"""
import os
import abc
import contextlib
import multiprocessing
from pathlib import Path
from .types import AppEnv
from lazyops.libs.proxyobj import proxied
from typing import Optional, Dict, Any, List, Union, Type, Callable, TypeVar, Literal, overload, TYPE_CHECKING

if TYPE_CHECKING:
    import jinja2
    from pydantic import BaseModel
    from kvdb import KVDBSession, PersistentDict
    from kvdb.tasks.base import BaseTaskWorker, TaskWorkerT
    from lazyops.utils.logs import Logger
    from lazyops.libs.fastapi_utils.types.persistence import TemporaryData
    from lazyops.libs.fastapi_utils.types.state import AppState
    from .base import AppSettings
    from ..clients import ClientTypes
    from ..sql.database.types import ObjectCRUD


    AppSettingsT = TypeVar('AppSettingsT', bound = AppSettings)
    ComponentSchemaT = TypeVar('ComponentSchemaT', BaseModel, ObjectCRUD)

class ApplicationContext(abc.ABC):
    """
    Contains the app context
    """

    def __init__(
        self,
        module_name: str,
        ingress_domain: str,
        ingress_base: Optional[str] = None,
        config_path: Optional[Path] = None,
        assets_dir: Optional[str] = 'assets',
        **kwargs,
        # global_ctx: Optional['GlobalContextClass'] = None,
    ):
        """
        Creates the app context
        """
        self.module_name = module_name
        self.app_env_var_name = self.module_name.replace(".", "-").replace("_", "-").lower()
        self.env_var_name = self.module_name.replace(".", "_").upper()
        self.is_app_module = '__main__' not in self.module_name
        self.import_assets: Callable[..., Path] = None
        self.get_assets: Callable[..., Path] = None
        self.ingress_base = ingress_base or self.module_name
        self.ingress_domain = ingress_domain
        if self.is_app_module:
            from lazyops.utils.assets import get_module_assets_path
            from lazyops.utils.assets import create_get_assets_wrapper, create_import_assets_wrapper
            if config_path is None:
                try:
                    config_path = get_module_assets_path(self.module_name, 'configs')
                except Exception as e:
                    config_path = Path.cwd()
            self.import_assets = create_import_assets_wrapper(self.module_name, assets_dir)
            self.get_assets = create_get_assets_wrapper(self.module_name, assets_dir)

        self.config_path = config_path
        # self.global_ctx = global_ctx

        self._app_env: Optional[AppEnv] = None
        self._temp_data: Optional['TemporaryData'] = None
        self._settings: Optional['AppSettings'] = None
        self._state: Optional['AppState'] = None
        self._logger: Optional['Logger'] = None
        self.on_close_funcs: List[Callable] = []

        self._kdb_sessions: Dict[str, 'KVDBSession'] = {}
        self._persistent_dicts: Dict[str, 'PersistentDict'] = {}
        self._persistent_dict_aliases: Dict[str, str] = {}

        # Client Registration and Hooks
        self._global_clients: Dict[str, str] = {}
        self._local_clients: Dict[str, str] = {}

        # Component Registration
        self._component_schema_registry: Dict[str, Union[Dict, str, Type]] = {}
        self._component_client_registry: Dict[str, Union[Dict, str, Any]] = {}

        self.include_kind_in_client_name: Optional[bool] = True
        self.include_kind_in_component_name: Optional[bool] = True
        
        self.client_kwarg_function: Optional[Union[Callable[..., Dict[str, Any]], str]] = None
        self.client_hook_function: Optional[Union[Callable, str]] = None

        self.component_client_hook_function: Optional[Union[Callable, str]] = None

        # Jinja2 Contexts
        self.j2_ctxs: Dict[str, 'jinja2.Environment'] = {}
        self.j2_actxs: Dict[str, 'jinja2.Environment'] = {}
    
        # Temp State
        self.temp_state: Dict[str, Any] = {}

    @property
    def disable_enforce_env(self) -> bool:
        """
        Returns whether to disable the enforce env
        """
        return os.getenv('DISABLE_ENFORCE_ENV', 'false').lower() in {'true', '1'}

    @property
    def app_env(self) -> AppEnv:
        """
        Retrieves the app environment
        """
        if self._app_env is None: self._app_env = self.get_app_env()
        return self._app_env
    
    @property
    def temp_data(self) -> 'TemporaryData':
        """
        Retrieves the temporary data
        """
        if self._temp_data is None:
            from lazyops.libs.fastapi_utils.types.persistence import TemporaryData
            self._temp_data = TemporaryData.from_module(self.module_name)
        return self._temp_data
    
    @property
    def settings(self) -> 'AppSettings':
        """
        Returns the settings
        """
        if self._settings is None:
            from .lazy import get_module_settings
            self._settings = get_module_settings(self.module_name)
        return self._settings
    
    @property
    def state(self) -> 'AppState':
        """
        Returns the state
        """
        if self._state is None:
            from lazyops.libs.fastapi_utils.types.state import AppState
            self._state = AppState()
            self._state.bind_settings(self.settings)
        return self._state
    

    @property
    def logger(self) -> 'Logger':
        """
        Returns the logger
        """
        if self._logger is None:
            from lazyops.utils.logs import logger as _logger
            self._logger = _logger
        return self._logger

    
    @property
    def is_leader_process(self) -> bool:
        """
        Returns if this is the leader process
        """
        return self.state.is_leader_process or multiprocessing.parent_process() is None
    
    @property
    def is_primary_server_process(self) -> bool:
        """
        Returns if this is the primary server process
        """
        return self.state.is_primary_server_process or multiprocessing.parent_process() is None
    

    def register_settings(self, settings: 'AppSettingsT') -> None:
        """
        Registers the settings
        """
        from .lazy import register_module_settings
        register_module_settings(self.module_name, settings)
        self._settings = settings

    """
    App Environments and Ingress
    """

    def get_app_env(self, module_name: Optional[str] = None) -> AppEnv:
        """
        Retrieves the app environment
        """
        module_name = module_name or self.env_var_name
        for key in {
            "SERVER_ENV",
            f"{module_name}_ENV",
            "APP_ENV",
            "ENVIRONMENT",
        }:
            if env_value := os.getenv(key):
                return AppEnv.from_env(env_value)

        # self.logger.info(f'Using: {self.app_env_var_name}')

        from lazyops.utils.system import is_in_kubernetes, get_host_name
        if is_in_kubernetes():
            # Name should be
            # scout-<service>-<index>
            # or 
            # scout-<service>-<env>-<index>

            # if the app name has a - in it, need to handle it differently
            # my-app-server-<index>
            # my-app-server-<env>-<index>
            hn = get_host_name()
            src_hn = hn.replace(self.app_env_var_name, '').strip()
            # if hn.count('-') 
            try:
                parts = src_hn.split('-')
                return AppEnv.from_env(parts[1]) if len(parts) > 2 else AppEnv.PRODUCTION
            except Exception as e:
                # self.logger.error(f'Error getting app env from hostname: {e}')
                return AppEnv.from_hostname(hn)
                # parts = get_host_name().split("-")
                # return AppEnv.from_env(parts[2]) if len(parts) > 3 else AppEnv.PRODUCTION

        return AppEnv.LOCAL
    

    def get_app_ingress(
        self,
        app_host: Optional[str] = None,
        app_port: Optional[int] = None,
    ) -> str:
        """
        Retrieves the app ingress
        """
        if self.app_env.is_local:
            app_host = app_host or "localhost"
            app_port = app_port or 8080
            return f"http://{app_host}:{app_port}"
        if self.app_env == AppEnv.DEVELOPMENT:
            return f"https://{self.ingress_base}-develop.{self.ingress_domain}"
        if self.app_env == AppEnv.STAGING:
            return f"https://{self.ingress_base}-staging.{self.ingress_domain}"
        if self.app_env == AppEnv.PRODUCTION:
            return f"https://{self.ingress_base}.{self.ingress_domain}"
        raise ValueError(f"Invalid app environment: {self.app_env}")

    def get_app_env_file(
        self,
        name: Optional[str] = None, 
        required: Optional[bool] = False, 
        allow_default: Optional[bool] = True,
        configs_path: Optional[Path] = None,
    ) -> Optional[Path]:
        """
        Retrieves the app environment file

        Only valid for local/dev environments
        """
        app_env = self.get_app_env()
        is_local_env = app_env in [
            AppEnv.LOCAL,
            AppEnv.DEVELOPMENT,
            AppEnv.CICD,
        ] or os.environ.get('DISABLE_ENFORCE_ENV', 'false').lower() == 'true'
        configs_path = configs_path or self.config_path
        envs_path = configs_path.joinpath('envs')
        if name is not None:
            if envs_path.joinpath(f'{name}-{app_env.name}.env').exists():
                return envs_path.joinpath(f'{name}-{app_env.name}.env')
            if envs_path.joinpath(f'{name}.env').exists():
                return envs_path.joinpath(f'{name}.env')
            if required and not is_local_env: raise ValueError(f"Invalid app environment file: {name}")
        env_path = envs_path.joinpath(f'{app_env.name}.env')
        if env_path.exists(): return env_path
        if (is_local_env or not required) and allow_default:
            return envs_path.joinpath('default.env')
        return None


    def get_app_default_file(
        self,
        name: Optional[str] = None, 
        required: Optional[bool] = False, 
        suffix: Optional[str] = None,
        configs_path: Optional[Path] = None,
    ) -> Optional[Path]:
        """
        Retrieves the app environment file

        Only valid for local/dev environments
        """
        app_env = self.get_app_env()
        configs_path = configs_path or self.config_path
        defaults_path = configs_path.joinpath('defaults')
        suffix = suffix or 'json'
        suffix = suffix.lstrip('.')
        if name is not None:
            if defaults_path.joinpath(f'{name}-{app_env.name}.{suffix}').exists():
                return defaults_path.joinpath(f'{name}-{app_env.name}.{suffix}')
            if defaults_path.joinpath(f'{name}.{suffix}').exists():
                return defaults_path.joinpath(f'{name}.{suffix}')
            if required: raise ValueError(f"Invalid app environment file: {name}")
        env_path = defaults_path.joinpath(f'{app_env.name}.{suffix}')
        if env_path.exists(): return env_path
        default_path = defaults_path.joinpath(f'default.{suffix}')
        return default_path if default_path.exists() else None
    
    """
    Migrate from lazy utils
    """

    def get_kdb_session(
        self,
        name: Optional[str] = None,
        serializer: Optional[str] = 'json',
        **kwargs,
    ) -> 'KVDBSession':
        """
        Retrieves or Initializes a KVDB Session
        """
        name = name or self.module_name
        if name not in self._kdb_sessions:
            from kvdb import KVDBClient
            self._kdb_sessions[name] = KVDBClient.get_session(
                name = name,
                serializer = serializer,
                **kwargs,
            )
        return self._kdb_sessions[name]
    

    def get_persistent_dict(
        self,
        base_key: str,
        expiration: Optional[int] = None,
        aliases: Optional[List[str]] = None,
        **kwargs,
    ) -> 'PersistentDict':
        """
        Lazily initializes a persistent dict
        """
        if base_key not in self._persistent_dicts and base_key not in self._persistent_dict_aliases:
            url = kwargs.pop('url', None)
            session = self.get_kdb_session('persistence', serializer = None, url = url)
            self._persistent_dicts[base_key] = session.create_persistence(
                base_key = base_key,
                expiration = expiration,
                **kwargs,
            )
            if aliases:
                for alias in aliases:
                    self._persistent_dict_aliases[alias] = base_key
        elif base_key in self._persistent_dict_aliases:
            base_key = self._persistent_dict_aliases[base_key]
        return self._persistent_dicts[base_key]

    def update_client_registry(
        self,
        clients: Dict[str, str],
        kind: Optional[str] = 'client',
        state: Literal['global', 'local'] = 'global',
        include_kind: Optional[bool] = None,
    ) -> None:
        """
        Updates the client registry
        """
        from lazyops.libs.fastapi_utils.state.registry import update_client_registry_mapping
        include_kind = include_kind if include_kind is not None else self.include_kind_in_client_name
        if include_kind:
            prefix = f'{self.module_name}.{kind}' if kind else self.module_name
        else:
            prefix = self.module_name
        
        client_mapping = {
            (f'{prefix}.{k}' if prefix not in k else k): v for k, v in clients.items()
        }
        update_client_registry_mapping(client_mapping)
        if state == 'global':
            self._global_clients.update(clients)
        else:
            self._local_clients.update(clients)

    def get_client_registry_mapping(self) -> Dict[str, str]:
        """
        Retrieves the client registry mapping
        """
        return {
            **self._global_clients,
            **self._local_clients,
        }

    def register_client(
        self,
        client: 'ClientTypes',
        kind: Optional[str] = 'client',
        include_kind: Optional[bool] = None,
    ) -> None:
        """
        Registers a client
        """
        from lazyops.libs.fastapi_utils.state.registry import register_client
        include_kind = include_kind if include_kind is not None else self.include_kind_in_client_name
        if include_kind:
            kind = kind or getattr(client, 'kind', 'client')
            prefix = f'{self.module_name}.{kind}'
        else:
            prefix = self.module_name
        name = f'{prefix}.{client.name}'
        return register_client(client, name)
    

    def get_client(
        self, 
        name: str,
        state: Optional[Literal['global', 'local']] = None,
        kind: Optional[str] = 'client',
        include_kind: Optional[bool] = None,
        **kwargs,
    ) -> 'ClientTypes':
        """
        Retrieves a client
        """
        if state is None:
            state = 'global' if name in self._global_clients else 'local'
        
        from lazyops.libs.fastapi_utils.state.registry import (
            get_client as _get_client, 
            get_global_client as _get_global_client, 
        )
        method = _get_global_client if state == 'global' else _get_client
        include_kind = include_kind if include_kind is not None else self.include_kind_in_client_name
        if include_kind:
            prefix = f'{self.module_name}.{kind}' if kind else self.module_name
        else:
            prefix = self.module_name
        client_name = name
        if prefix not in client_name:
            client_name = f'{prefix}.{client_name}'
        if self.client_kwarg_function:
            if isinstance(self.client_kwarg_function, str):
                from lazyops.utils.lazy import lazy_import
                self.client_kwarg_function = lazy_import(self.client_kwarg_function)
            kwargs = self.client_kwarg_function(name, **kwargs)
        client = method(client_name, **kwargs)
        if self.client_hook_function:
            if isinstance(self.client_hook_function, str):
                from lazyops.utils.lazy import lazy_import
                self.client_hook_function = lazy_import(self.client_hook_function)
            self.client_hook_function(name, client)
        return client


    def update_component_client_registry(
        self,
        components: Dict[str, Union[str, Dict]],
        kind: Optional[str] = None,
        include_kind: Optional[bool] = None,
    ) -> None:
        """
        Updates the component client registry
        """
        from lazyops.libs.fastapi_utils.state.registry import update_client_registry_mapping
        from lazyops.libs.abcs.utils.helpers import flatten_dict_value 
        include_kind = include_kind if include_kind is not None else self.include_kind_in_component_name
        if include_kind:
            prefix = f'{self.module_name}.{kind}' if kind else self.module_name
        else:
            prefix = self.module_name

        mapping = flatten_dict_value(components, prefix)
        update_client_registry_mapping(mapping)
        self._component_client_registry.update(components)

    def get_component_client(
        self,
        name: str,
        *parts: str,
        kind: Optional[str] = None,
        include_kind: Optional[bool] = None,
    ) -> 'ClientTypes':
        """
        Gets a component client
        """
        from lazyops.libs.fastapi_utils.state.registry import get_client
        include_kind = include_kind if include_kind is not None else self.include_kind_in_component_name
        if include_kind: client_name = f'{self.module_name}.{kind}' if kind else self.module_name
        else: client_name = self.module_name
        if parts:
            parts = '.'.join(parts)
            client_name = f'{client_name}.{parts}'
        client_name = f'{client_name}.{name}'
        return get_client(client_name)
    
    def register_component_client(
        self,
        client: 'ClientTypes',
        *parts: str,
        kind: Optional[str] = None,
        include_kind: Optional[bool] = None,
    ):
        """
        Registers a component client
        """
        from lazyops.libs.fastapi_utils.state.registry import register_client
        include_kind = include_kind if include_kind is not None else self.include_kind_in_component_name
        kind = kind or getattr(client, 'kind', None)
        if include_kind: prefix = f'{self.module_name}.{kind}' if kind else self.module_name
        else: prefix = self.module_name
        if parts:
            parts = '.'.join(parts)
            prefix = f'{prefix}.{parts}'
        client_name = f'{prefix}.{client.name}'
        return register_client(client, client_name)
        
    def update_component_schema_registry(
        self,
        components: Dict[str, Union[str, Dict]],
        kind: Optional[str] = None,
        include_kind: Optional[bool] = None,
    ) -> None:
        """
        Updates the component schema registry
        """
        from lazyops.libs.abcs.utils.helpers import flatten_dict_value 
        include_kind = include_kind if include_kind is not None else self.include_kind_in_component_name
        if include_kind:
            prefix = f'{self.module_name}.{kind}' if kind else self.module_name
        else:
            prefix = self.module_name
        
        mapping = flatten_dict_value(components, prefix)
        self._component_schema_registry.update(mapping)

    def get_component_schema(
        self,
        name: str,
        *parts: str,
        kind: Optional[str] = None,
        include_kind: Optional[bool] = None,
    ) -> Type['ComponentSchemaT']:
        """
        Gets a component schema
        """
        include_kind = include_kind if include_kind is not None else self.include_kind_in_component_name
        if include_kind: schema_name = f'{self.module_name}.{kind}' if kind else self.module_name
        else: schema_name = self.module_name
        if parts:
            parts = '.'.join(parts)
            schema_name = f'{schema_name}.{parts}'
        schema_name = f'{schema_name}.{name}'
        if schema_name not in self._component_schema_registry:
            raise ValueError(f"Invalid component schema: {schema_name}")
        schema = self._component_schema_registry[schema_name]
        if isinstance(schema, str):
            from lazyops.utils.lazy import lazy_import
            schema = lazy_import(schema)
        return schema
    
    def get_worker(
        self,
        name: Optional[str] = 'worker',
        state: Optional[Literal['global', 'local']] = None,
        kind: Optional[str] = 'client',
        include_kind: Optional[bool] = None,
        **kwargs,
    ) -> 'TaskWorkerT':
        """
        Retrieves the Worker
        """
        return self.get_client(name, state = state, kind = kind, include_kind = include_kind, **kwargs)
        

    @overload
    def get_j2_ctx(
        self,
        path: Union[str, Path],
        name: Optional[str] = None,
        enable_async: Optional[bool] = False,
        comment_start_string: Optional[str] = None,
        comment_end_string: Optional[str] = None,
        **kwargs,
    ) -> 'jinja2.Environment':
        """
        Creates a jinja2 context

        path can be relative to the module_dir or absolute
        """
        ...


    def get_j2_ctx(
        self,
        
        path: Union[str, Path],
        name: Optional[str] = None,
        enable_async: Optional[bool] = False,
        **kwargs,
    ) -> 'jinja2.Environment':
        """
        Creates a jinja2 context

        path can be relative to the module_dir or absolute
        """
        base_ctx = self.j2_actxs if enable_async else self.j2_ctxs
        name = name or self.module_name
        if name not in base_ctx:
            import jinja2
            if isinstance(path, str):
                if path.startswith('/'):
                    path = Path(path)
                else:
                    path = self.settings.module_path.joinpath(path)
            self.logger.info(f"Jinja2 Path: {path} for {name}")
            base_ctx[name] = jinja2.Environment(
                loader = jinja2.FileSystemLoader(path),
                enable_async = enable_async,
                **kwargs,
            )
        return base_ctx[name]

    def lazy_import(self, dotted_path: str, **kwargs) -> Any:
        """
        Lazy imports a module
        """
        from lazyops.utils.lazy import lazy_import
        return lazy_import(dotted_path, **kwargs)


# class ApplicationContextManagerClass(abc.ABC):


@proxied
class ApplicationContextManager(abc.ABC):

    """
    The context manager for the app context
    """

    def __init__(self, *args, **kwargs):

        """
        Creates the context manager
        """
        self.ctxs: Dict[str, ApplicationContext] = {}
        self._global_ctx = None
    
    @property
    def global_ctx(self): # -> 'GlobalContextClass':
        """
        Returns the global context
        """
        if self._global_ctx is None:
            from ..state import GlobalContext
            self._global_ctx = GlobalContext
        return self._global_ctx

    def init_ctx(self, module_name: str, *args, **kwargs) -> ApplicationContext:
        """
        Initializes the app context
        """
        if module_name not in self.ctxs:
            self.ctxs[module_name] = ApplicationContext(module_name, *args, global_ctx = self.global_ctx, **kwargs)
        return self.ctxs[module_name]
    
    def get_ctx(self, module_name: str, *args, **kwargs) -> ApplicationContext:
        """
        Retrieves the app context
        """
        if module_name not in self.ctxs:
            return self.init_ctx(module_name, *args, **kwargs)
        return self.ctxs[module_name]
    
    def __getitem__(self, module_name: str) -> ApplicationContext:
        """
        Retrieves the app context
        """
        return self.get_ctx(module_name)
    


# ApplicationContextManager: 'ApplicationContextManagerClass' = ProxyObject(
#     ApplicationContextManagerClass
# )


from __future__ import annotations

import abc
import atexit
import sqlite3
import pathlib
import filelock
import contextlib
from pydantic import BaseModel, Field, model_validator, model_serializer
from lazyops.imports._aiosqlite import resolve_aiosqlite
from lazyops.utils.lazy import lazy_import
from lazyops.utils import logger, Timer
from typing import Optional, List, Dict, Any, Union, Type, Tuple, TypeVar, AsyncGenerator, overload, TYPE_CHECKING

if TYPE_CHECKING:
    
    # import aiosqlite
    from lazyops.libs import aiosqlite
    from .mixins import SQLResultT, SQLiteModelMixin
    

    # SchemaT = Union[Type[SQLResultT], str]
    # SchemaListT = List[Union[Type[SQLResultT], str]]

    SchemaT = Union[Type[SQLiteModelMixin], str]
    SchemaListT = List[Union[Type[SQLiteModelMixin], str]]

    ModelT = TypeVar('ModelT', bound = SQLiteModelMixin)

    # SchemaTypeT = TypeVar('SchemaTypeT', bound = SQLiteModelMixin)

    # SchemaTypeT = SQLResultT



class SQLiteIndex(abc.ABC):
    """
    The SQLite Index

    This is a base class for all SQLite Indexes. It provides common functionality for indexes, such as creating and updating indexes, and searching for items.

    Attributes:
        sql_data_path (pathlib.Path): The path to the SQLite database file.
        tablename (str): The name of the table in the database.
        schema (str): The schema of the table.
        schemas (Dict[str, Union[str, List[str], Dict[str, str]]]): The schemas of the table.
        ephemeral (bool): Whether the index is ephemeral or not.
        force_reindex (bool): Whether to force reindexing the index.
        auto_set_schema (bool): Whether to automatically set the schema of the index.
        enable_internal_index (bool): Whether to enable the internal index.
    """

    sql_data_path: Optional[pathlib.Path] = None

    # Can define the schema to use as a single type or multiple types
    tablename: Optional[str] = None
    schema: Optional['SchemaT'] = None
    schemas: Optional[Dict[str, Union['SchemaT', 'SchemaListT']]] = None

    ephemeral: Optional[bool] = False
    force_reindex: Optional[bool] = False
    auto_set_schema: Optional[bool] = False
    enable_internal_index: Optional[bool] = False

    def __init__(
        self,
        sql_data_path: Optional[pathlib.Path] = None,
        tablename: Optional[str] = None,
        schema: Optional['SchemaT'] = None,
        schemas: Optional[Dict[str, Union['SchemaT', 'SchemaListT']]] = None,
        ephemeral: Optional[bool] = None,
        force_reindex: Optional[bool] = None,
        auto_set_schema: Optional[bool] = None,
        enable_internal_index: Optional[bool] = None,
        **kwargs,
    ):
        """
        Initializes the SQLite Index
        """
        self._initialized: bool = False
        super().__init__(**kwargs)
        self.cls_init(
            sql_data_path = sql_data_path, 
            tablename = tablename, 
            schema = schema, 
            schemas = schemas, 
            ephemeral = ephemeral, 
            force_reindex = force_reindex, 
            auto_set_schema = auto_set_schema,
            enable_internal_index = enable_internal_index,
            **kwargs
        )

    def cls_init(
        self, 
        sql_data_path: Optional[pathlib.Path] = None,
        tablename: Optional[str] = None,
        schema: Optional['SchemaT'] = None,
        schemas: Optional[Dict[str, Union['SchemaT', 'SchemaListT']]] = None,
        ephemeral: Optional[bool] = None,
        force_reindex: Optional[bool] = None,
        auto_set_schema: Optional[bool] = None,
        enable_internal_index: Optional[bool] = None,
        **kwargs
    ):  # sourcery skip: low-code-quality
        """
        For compatibility with `BaseGlobalClient`
        """
        if self._initialized: return
        if sql_data_path is not None: self.sql_data_path = sql_data_path
        if tablename is not None: self.tablename = tablename
        if schema is not None: self.schema = schema
        if schemas is not None: self.schemas = schemas
        if enable_internal_index is not None: self.enable_internal_index = enable_internal_index
        self.schemas = self.schemas or {}
        self.internal_index: Dict[str, Dict[str, 'SQLResultT']] = {}
        # self.schemas = schemas if schemas is not None else {}
        if self.tablename and self.schema:
            self.schemas[self.tablename] = self.schema
        
        if ephemeral is not None: self.ephemeral = ephemeral
        if force_reindex is not None: self.force_reindex = force_reindex
        if auto_set_schema is not None: self.auto_set_schema = auto_set_schema
        if self.schema is not None:
            if isinstance(self.schema, str): self.schema = lazy_import(self.schema)

        if self.schemas is not None:
            for key, value in self.schemas.items():
                if isinstance(value, list):
                    s_values = []
                    for v in value:
                        if isinstance(v, str): v = lazy_import(v)
                        s_values.append(v)
                    value = s_values
                    self.schemas[key] = s_values
                elif isinstance(value, str): self.schemas[key] = lazy_import(value)
        

        self.default_tablename = self.tablename or list(self.schemas.keys())[0]
        self._conn: Optional['sqlite3.Connection'] = None
        self._aconn: Optional['aiosqlite.Connection'] = None

        self._conn_initialized: Optional[bool] = None
        self._aconn_initialized: Optional[bool] = None

        self._schemas_initialized: Optional[bool] = None
        self._sql_lock: Optional[filelock.SoftFileLock] = None
        self.sql_lock_path: Optional[pathlib.Path] = None
        if self.sql_data_path: 
            if self.sql_data_path.is_dir() and self.tablename:
                self.sql_data_path.mkdir(parents = True, exist_ok = True)
                self.sql_data_path = self.sql_data_path.joinpath(f'{self.tablename}.db')
            self.sql_lock_path = self.sql_data_path.with_suffix('.lock')
        self._initialized = True
    
    @property
    def sql_lock(self) -> filelock.SoftFileLock:
        """
        Returns the SQL Lock
        """
        if self._sql_lock is None:
            self._sql_lock = filelock.SoftFileLock(
                self.sql_lock_path.as_posix(),
                timeout = 0,
                thread_local = False,
            )
            if self.ephemeral: atexit.register(self.cleanup_on_exit)
            atexit.register(self.save_on_exit)
        return self._sql_lock

    def save_on_exit(self):
        """
        Saves the index on exit
        """
        self.conn.close()
        # self.conn.backup()


    def cleanup_on_exit(self):
        """
        Cleans up the index on exit
        """
        if self.sql_lock_path.exists():
            logger.info(f'Removing SQLite Lock: {self.sql_lock_path}', prefix = '|g|DB|e|', colored = True)
            try:
                self.sql_lock.release()
                self.sql_lock_path.unlink()
            except Exception as e:
                logger.error(f'Error cleaning up SQLite Lock: {e}', prefix = '|g|DB|e|', colored = True)
        
        if self.sql_data_path.exists():
            logger.info(f'Removing SQLite Data: {self.sql_data_path}', prefix = '|g|DB|e|', colored = True)
            try:
                self.sql_data_path.unlink()
            except Exception as e:
                logger.error(f'Error cleaning up SQLite Data: {e}', prefix = '|g|DB|e|', colored = True)
        
    def get_index_schemas(self) -> Dict[str, Union[Type['SQLiteModelMixin'], List[Type['SQLiteModelMixin']]]]:
        """
        Gets the index schemas
        """
        return (
            self.schemas or 
            {
                self.tablename: self.schema,
            }
        )

    def conn_prehook(self, **kwargs):
        """
        This is called before the connection is created
        """
        pass


    def conn_posthook(self, **kwargs):
        """
        This is called after the connection is created
        """
        pass

    async def aconn_prehook(self, **kwargs):
        """
        This is called before the async connection is created
        """
        pass

    async def aconn_posthook(self, **kwargs):
        """
        This is called after the async connection is created
        """
        pass

    def init_schemas(self, conn: 'sqlite3.Connection', exists: Optional[bool] = None) -> None:
        """
        Initializes the schemas
        """
        if exists is None: exists = self.sql_data_path.exists()
        if not self.sql_lock.is_locked:
            with self.sql_lock.acquire(10.0):
                index_schemas = self.get_index_schemas()
                for tablename, schema in index_schemas.items():
                    if not isinstance(schema, list): schema = [schema]
                    for s in schema:
                        s.execute_sql_init(
                            conn, 
                            tablename = tablename, 
                            skip_index = (exists and not self.force_reindex), 
                            auto_set = self.auto_set_schema
                        )
        self._schemas_initialized = True


    def init_db_schemas(self) -> None:
        """
        Initializes the database schemas
        """
        if self._schemas_initialized: return
        self.conn_prehook()
        conn = sqlite3.connect(self.sql_data_path.as_posix(), check_same_thread = False)
        self.init_schemas(conn)
        self.conn_posthook()


    @property
    def conn(self) -> 'sqlite3.Connection':
        """
        Returns the connection
        """
        if self._conn is None:
            self.conn_prehook()
            exists = self.sql_data_path.exists()
            import sqlite3
            self._conn = sqlite3.connect(self.sql_data_path.as_posix(), check_same_thread = False)
            if not self._schemas_initialized:
                self.init_schemas(self._conn, exists = exists)
        return self._conn
    
    async def ainit_schemas(self, conn: 'aiosqlite.Connection') -> None:
        """
        Initializes the schemas
        """
        index_schemas = self.get_index_schemas()
        exists = self.sql_data_path.exists()
        with self.sql_lock.acquire(10.0):
            for tablename, schema in index_schemas.items():
                if not isinstance(schema, list): schema = [schema]
                for s in schema:
                    await s.aexecute_sql_init(
                        conn, 
                        tablename = tablename, 
                        skip_index = (exists and not self.force_reindex), 
                        auto_set = self.auto_set_schema
                    )
        # logger.info(f'Completed SQL initialization in {t.total_s}')
        self._schemas_initialized = True
    
    async def ainit_db_schemas(self):
        """
        Initializes the database schemas
        """
        if self._schemas_initialized: return
        from lazyops.libs import aiosqlite
        await self.aconn_prehook()
        conn = await aiosqlite.connect(self.sql_data_path.as_posix())
        await self.ainit_schemas(conn)


    @contextlib.asynccontextmanager
    async def aconn(self) -> AsyncGenerator['aiosqlite.Connection', None]:
        """
        Returns the async connection
        """
        if self._aconn is None:
            from lazyops.libs import aiosqlite
            await self.aconn_prehook()
            self._aconn = await aiosqlite.connect(self.sql_data_path.as_posix(), limit = 10)
            if not self._schemas_initialized:
                await self.ainit_schemas(self._aconn)
        try:
            yield self._aconn
        except Exception as e:
            logger.trace('Error in async connection', e)
            raise e
        finally:
            await self.aconn_posthook()

        # resolve_aiosqlite(True)
        # import aiosqlite
        # await self.aconn_prehook()
        # async with aiosqlite.connect(self.sql_data_path.as_posix()) as conn:
        #     if not self._aconn_initialized:
        #         await self.ainit_schemas(conn)
        #     try:
        #         yield conn
        #     except Exception as e:
        #         logger.trace('Error in async connection', e)
        #     finally:
        #         await self.aconn_posthook()
        # from lazyops.libs import aiosqlite
        # await self.aconn_prehook()
        # conn = await aiosqlite.connect(self.sql_data_path.as_posix())
        # if not self._aconn_initialized:
        #     await self.ainit_schemas(conn)
        # try:
        #     yield conn
        # except Exception as e:
        #     logger.trace('Error in async connection', e)
        # finally:
        #     await self.aconn_posthook()




    
    def get_schema(
        self,
        schema_name: Optional[str] = None,
        tablename: Optional[str] = None,
    ) -> 'SchemaT':
        """
        Gets the schema by name
        """
        if tablename is None: tablename = self.tablename
        schemas = self.schemas[tablename]
        if not isinstance(schemas, list): return schemas
        if schema_name is None: return schemas[0]
        for schema in schemas:
            if schema_name in schema.__name__:
                return schema
        raise ValueError(f'Schema {schema_name} not found in {tablename}')
    
    def format_query(
        self,
        query: Optional[str] = None,
        **kwargs,
    ) -> Tuple[Optional[str], Dict[str, Any]]:
        """
        Helper Function to Preformat the query
        """
        return query, kwargs

    @overload
    def search_one(
        self,
        query: Optional[str] = None,
        skip: Optional[int] = None,
        tablename: Optional[str] = None,
        schema_name: Optional[str] = None,
        **kwargs,
    ) -> 'SQLiteModelMixin':
        """
        Executes the sql search for one item
        """
        ...

    @overload
    def search_one(
        self,
        query: Optional[str] = None,
        skip: Optional[int] = None,
        return_id_only: Optional[bool] = True,
        tablename: Optional[str] = None,
        schema_name: Optional[str] = None,
        **kwargs,
    ) -> str:
        """
        Returns the id of the first item that matches the query
        """
        ...


    def search_one(
        self,
        query: Optional[str] = None,
        skip: Optional[int] = None,
        return_id_only: Optional[bool] = None,
        return_fields: Optional[List[str]] = None,
        tablename: Optional[str] = None,
        schema_name: Optional[str] = None,
        cast_to_object: Optional[bool] = None,
        **kwargs,
    ) -> Optional[Union['SQLResultT', str]]:
        """
        Executes the sql search for one item
        """
        tablename = tablename or self.default_tablename
        schema = self.get_schema(tablename = tablename, schema_name = schema_name)
        if self.enable_internal_index and (return_id_only is None and return_fields is None) and self.internal_index.get(tablename):
            return_id_only = True
            if cast_to_object is None: cast_to_object = True
        query, kwargs = self.format_query(query, **kwargs)
        result = schema.search_one(
            conn = self.conn,
            query = query, 
            skip = skip, 
            return_id_only = return_id_only, 
            return_fields = return_fields,
            tablename = tablename,
            **kwargs
        )
        if self.enable_internal_index and result and cast_to_object:
            return self.populate_item_from_index(tablename, result)
        return result

    async def asearch_one(
        self,
        query: Optional[str] = None,
        skip: Optional[int] = None,
        return_id_only: Optional[bool] = None,
        return_fields: Optional[List[str]] = None,
        tablename: Optional[str] = None,
        schema_name: Optional[str] = None,
        cast_to_object: Optional[bool] = None,
        **kwargs,
    ) -> Optional[Union['SQLiteModelMixin', str]]:
        """
        Executes the sql search for one item
        """
        tablename = tablename or self.default_tablename
        schema = self.get_schema(tablename = tablename, schema_name = schema_name)
        if self.enable_internal_index and (return_id_only is None and return_fields is None) and self.internal_index.get(tablename):
            return_id_only = True
            cast_to_object = True
        query, kwargs = self.format_query(query, **kwargs)
        async with self.aconn() as conn:
            result = await schema.asearch_one(
                conn = conn,
                query = query, 
                skip = skip, 
                return_id_only = return_id_only, 
                return_fields = return_fields,
                tablename = tablename,
                **kwargs
            )
        if self.enable_internal_index and result and cast_to_object:
            return self.populate_item_from_index(tablename, result)
        return result

    def search_many(
        self,
        query: Optional[str] = None,
        limit: Optional[int] = None,
        skip: Optional[int] = None,
        return_id_only: Optional[bool] = None,
        return_fields: Optional[List[str]] = None,
        tablename: Optional[str] = None,
        schema_name: Optional[str] = None,
        cast_to_object: Optional[bool] = None,
        **kwargs,
    ) -> List[Union['SQLiteModelMixin', str]]:
        """
        Executes the sql search for many items
        """
        tablename = tablename or self.default_tablename
        schema = self.get_schema(tablename = tablename, schema_name = schema_name)
        if self.enable_internal_index and (return_id_only is None and return_fields is None) and self.internal_index.get(tablename):
            return_id_only = True
            cast_to_object = True
        query, kwargs = self.format_query(query, **kwargs)
        results = schema.search_many(
            conn = self.conn,
            query = query, 
            limit = limit, 
            skip = skip, 
            return_id_only = return_id_only, 
            return_fields = return_fields,
            tablename = tablename,
            **kwargs
        )
        if self.enable_internal_index and results and cast_to_object:
            return [self.populate_item_from_index(tablename, result) for result in results]
        return results

    async def asearch_many(
        self,
        query: Optional[str] = None,
        limit: Optional[int] = None,
        skip: Optional[int] = None,
        return_id_only: Optional[bool] = None,
        return_fields: Optional[List[str]] = None,
        tablename: Optional[str] = None,
        schema_name: Optional[str] = None,
        cast_to_object: Optional[bool] = None,
        **kwargs,
    ) -> List[Union['SQLiteModelMixin', str]]:
        """
        Executes the sql search for many items
        """
        tablename = tablename or self.default_tablename
        schema = self.get_schema(tablename = tablename, schema_name = schema_name)
        if self.enable_internal_index and (return_id_only is None and return_fields is None) and self.internal_index.get(tablename):
            return_id_only = True
            cast_to_object = True
        query, kwargs = self.format_query(query, **kwargs)
        async with self.aconn() as conn:
            try:
                results = await schema.asearch_many(
                    conn = conn,
                    query = query, 
                    limit = limit, 
                    skip = skip, 
                    return_id_only = return_id_only, 
                    return_fields = return_fields,
                    tablename = tablename,
                    **kwargs
                )
            except Exception as e:
                logger.error(f'[{tablename}] Error Searching: {e}, {query}, {kwargs}')
                raise e
            
        if self.enable_internal_index and results and cast_to_object:
            return [self.populate_item_from_index(tablename, result) for result in results]
        return results
        

    def delete(
        self,
        item: 'SQLiteModelMixin',
        **kwargs,
    ) -> None:
        """
        Deletes the object from the database
        """
        item.sql_delete(self.conn)
        self.unindex_item(item)


    async def adelete(
        self,
        item: 'SQLiteModelMixin',
        **kwargs,
    ) -> None:
        """
        Deletes the object from the database
        """
        async with self.aconn() as conn:
            await item.asql_delete(conn)
        self.unindex_item(item)
    
    def unindex_item(
        self,
        item: 'SQLiteModelMixin',
        **kwargs,
    ) -> None:
        """
        Unindexes the item
        """
        if self.enable_internal_index:
            if item.sql_tablename in self.internal_index:
                del self.internal_index[item.sql_tablename][item.sql_pkey]

    def index_item(
        self,
        item: 'SQLiteModelMixin',
        **kwargs,
    ) -> None:
        """
        Indexes the item
        """
        if self.enable_internal_index:
            if item.sql_tablename not in self.internal_index:
                self.internal_index[item.sql_tablename] = {}
            self.internal_index[item.sql_tablename][item.sql_pkey] = item
    
    def populate_item_from_index(
        self,
        tablename: str,
        pkey: Union[str, int, Dict[str, Any], List[Union[str, int]]],
        **kwargs,
    ) -> Optional['SQLiteModelMixin']:
        """
        Populates the item from the index
        """
        if tablename not in self.internal_index: return None
        if isinstance(pkey, dict): pkey = list(pkey.values())
        if not isinstance(pkey, list): pkey = [pkey]
        return next(
            (
                self.internal_index[tablename][pk]
                for pk in pkey
                if pk is not None and pk in self.internal_index[tablename]
            ),
            None,
        )
        # if pkey not in self.internal_index[tablename]: return None
        # return self.internal_index[tablename][pkey]
    
    def index_items(
        self,
        items: List['SQLiteModelMixin'],
        **kwargs,
    ) -> None:
        """
        Indexes the items
        """
        if self.enable_internal_index:
            for item in items:
                self.index_item(item)
    
    def set(
        self,
        item: 'SQLiteModelMixin',
        skip_set: Optional[bool] = None,
        skip_index: Optional[bool] = None,
        **kwargs,
    ) -> None:
        """
        Saves the object to the database
        """
        if not skip_set: item.sql_set(self.conn)
        if not skip_index: self.index_item(item)
        # item.sql_set(self.conn)
        # self.index_item(item)
        
    async def aset(
        self,
        item: 'SQLiteModelMixin',
        skip_set: Optional[bool] = None,
        skip_index: Optional[bool] = None,
        **kwargs,
    ):
        """
        Saves the object to the database
        """
        async with self.aconn() as conn:
            if not skip_set: await item.asql_set(conn)
            if not skip_index: self.index_item(item)
            # await item.asql_set(conn)
            # self.index_item(item)

    def set_many(
        self,
        items: List['SQLiteModelMixin'],
        skip_index: Optional[bool] = None,
        **kwargs,
    ):
        """
        Saves the objects to the database
        """
        item_cls = items[0].__class__
        item_cls.sql_set_many(self.conn, items = items, **kwargs)
        if not skip_index: self.index_items(items)

    async def aset_many(
        self,
        items: List['SQLiteModelMixin'],
        skip_index: Optional[bool] = None,
        **kwargs,
    ):
        """
        Saves the objects to the database
        """
        item_cls = items[0].__class__
        async with self.aconn() as conn:
            await item_cls.asql_set_many(conn, items = items, **kwargs)
        if not skip_index: self.index_items(items)
        
    def reindex_from_internal_index(
        self,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        **kwargs,
    ) -> int:
        """
        Saves all items from the internal index to the DB
        """
        include = include or []
        exclude = exclude or []
        count = 0
        
        for tablename, index in self.internal_index.items():
            if (not include or tablename in include) and (not exclude or tablename not in exclude):
                items = list(index.values())
                if not items: continue
                self.set_many(items, skip_index = True, **kwargs)
                count += len(items)
        return count
    
    async def areindex_from_internal_index(
        self,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        **kwargs,
    ) -> int:
        """
        Saves all items from the internal index to the DB
        """
        include = include or []
        exclude = exclude or []
        count = 0
        t = Timer()
        for tablename, index in self.internal_index.items():
            if (not include or tablename in include) and (not exclude or tablename not in exclude):
                items = list(index.values())
                if not items: continue
                await self.aset_many(items, skip_index = True, **kwargs)
                count += len(items)
        logger.info(f'Completed Reindexing {count} items in {t.duration_s}')
        return count
        



    def __getitem__(self, name: str) -> Union['SQLiteModelMixin', Any]:
        """
        Gets the item from the db
        """
        if ':' in name: tablename, name = name.split(':', 1)
        else: tablename = self.default_tablename
        return self.search_one(tablename = tablename, query = name)
    

    def __add__(self, item: 'SQLiteModelMixin') -> 'SQLiteModelMixin':
        """
        Adds the item to the database
        """
        self.set(item)
        return item
    
    def __sub__(self, item: 'SQLiteModelMixin') -> 'SQLiteModelMixin':
        """
        Removes the item from the database
        """
        item.sql_delete(self.conn)
        return item
    
    def __iadd__(self, item: 'SQLiteModelMixin') -> 'SQLiteModelMixin':
        """
        Adds the item to the database
        """
        self.set(item)
        return item
    
    def __isub__(self, item: 'SQLiteModelMixin') -> 'SQLiteModelMixin':
        """
        Removes the item from the database
        """
        self.delete(item)
        return item
    




from __future__ import annotations

"""
Template Manager
"""

import abc
from pathlib import Path
from typing import Optional, Any, Dict, List, Union, Type, TYPE_CHECKING

if TYPE_CHECKING:
    import jinja2
    from lazyops.libs.abcs.configs.base import AppSettings


class SQLTemplates(abc.ABC):
    """
    SQL Template Manager
    """

    name: Optional[str] = None
    base_path: Union[str, Path] = None
    file_suffix: Optional[str] = '.sql'

    def __init__(
        self, 
        
        name: Optional[str] = None,
        base_path: Union[str, Path] = None, 
        file_suffix: Optional[str] = None,
        settings: Optional['AppSettings'] = None,
        module_name: Optional[str] = None,
    ):
        """
        Initializes the SQL Template Manager
        """
        if base_path: self.base_path = base_path
        if file_suffix: self.file_suffix = file_suffix
        if name is not None: self.name = name
        if settings is None:
            from lazyops.libs.abcs.configs.lazy import get_module_settings
            settings = get_module_settings(module_name or self.__module__.__name__)
        
        self.settings = settings
        if self.name is None:
            self.name = f'{self.settings.module_name}_sql'
        self._j2: Optional['jinja2.Environment'] = None
        self._aj2: Optional['jinja2.Environment'] = None

    @property
    def j2(self) -> 'jinja2.Environment':
        """
        Returns the jinja2 context
        """
        if self._j2 is None:
            self._j2 = self.settings.ctx.get_j2_ctx(
                path = self.base_path or 'assets/sql',
                name = self.name,
                enable_async = False,
                comment_start_string='/*',
                comment_end_string='*/',
            )
        return self._j2
    
    @property
    def aj2(self) -> 'jinja2.Environment':
        """
        Returns the jinja2 context
        """
        if self._aj2 is None:
            self._aj2 = self.settings.ctx.get_j2_ctx(
                path = self.base_path or 'assets/sql',
                name = self.name,
                enable_async = True,
                comment_start_string='/*',
                comment_end_string='*/',
            )
        return self._aj2
    
    def get_template(self, *parts: str) -> 'jinja2.Template':
        """
        Gets the template
        """
        if not parts[-1].endswith(self.file_suffix):
            parts = list(parts[:-1]) + [parts[-1] + self.file_suffix]
        return self.j2.get_template('/'.join(parts))
    
    def aget_template(self, *parts: str) -> 'jinja2.Template':
        """
        Gets the template
        """
        if not parts[-1].endswith(self.file_suffix):
            parts = list(parts[:-1]) + [parts[-1] + self.file_suffix]
        return self.aj2.get_template('/'.join(parts))
        

    def render(self, *parts: str, **kwargs) -> str:
        """
        Renders the template
        """
        return self.get_template(*parts).render(**kwargs)
    

    async def arender(self, *parts: str, **kwargs) -> str:
        """
        Renders the template
        """
        # template = self.aget_template(*parts)
        # template.
        return await self.aget_template(*parts).render_async(**kwargs)
    
    def __call__(
        self,
        *parts: str,
        enable_async: bool = False,
        **kwargs,
    ) -> str:
        """
        Renders the template
        """
        if enable_async:
            return self.arender(*parts, **kwargs)
        return self.render(*parts, **kwargs)
    
    def __getitem__(self, name: str) -> 'jinja2.Template':
        """
        Gets the template
        """
        parts = name.split('.')
        return self.get_template(*parts)
    
    @property
    def sql_path(self) -> Path:
        """
        Returns the SQL Path
        """
        if self.base_path and isinstance(self.base_path, Path):
            return self.base_path
        p = self.base_path or 'assets/sql'
        if isinstance(p, str):
            p = self.settings.module_path.joinpath(p)
        return p
    
    def get_migration_sqls(
        self,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        path: Optional[Union[str, Path]] = None,
    ) -> List[str]:
        """
        Returns the migration sqls
        """
        sqls = []
        path = path or 'migrations'
        if isinstance(path, str):
            path = self.sql_path.joinpath(path)
        if not path.exists(): return sqls
        for name in path.iterdir():
            if name.suffix != '.sql': continue
            name_stem = name.stem.split('.', 1)[0]
            if include and name_stem not in include: continue
            if exclude and name_stem in exclude: continue
            sqls.append(name_stem)
        return sqls









import random
import inspect
import backoff
import functools
import aiohttpx
from typing import Callable, List, Optional, TypeVar, TYPE_CHECKING

try:
    from async_lru import alru_cache
except ImportError:
    import sys
    if sys.version_info >= (3, 8):
        from typing import ParamSpec
    else:
        from typing_extensions import ParamSpec

    RT = TypeVar('RT')
    PT = ParamSpec('PT')

    def alru_cache(*args, **kwargs):
        def decorator(func: Callable[PT, RT]) -> Callable[PT, RT]:
            return func
        return decorator

        

def create_cachify_build_name_func(
    base_name: str,
    special_names: Optional[List[str]] = None,
    function_names: Optional[List[str]] = None,
    include_classname: Optional[bool] = True,
    include_http_methods: Optional[bool] = False,
) -> Callable[..., str]:
    """
    Creates a function that returns the build name
    """
    # from .logs import logger
    # logger.info(f"Creating build name function for {base_name}")

    def build_name_func(func: Callable, *args, **kwargs) -> str:
        """
        Build the name function
        """
        nonlocal base_name
        func = inspect.unwrap(func)
        func_name = func.__qualname__ if include_classname else func.__name__

        if special_names:
            for name in special_names:
                if name in func_name.lower():
                    func_name = func_name.replace(f'a{name}_', '').replace(f'{name}_', '').replace('__', '_')
                    return f"{base_name}.{name}_{func_name}"
        
        if function_names:
            for name in function_names:
                if name in func_name.lower(): return f"{base_name}.{func_name}"
        
        if include_http_methods:
            for method in {'get', 'post', 'put', 'delete'}:
                if method in func_name.lower(): return f"{base_name}.{method}"

        # logger.info(f"Failed to find a special name for {base_name}.{func_name}")
        return f"{base_name}.{func_name}"
    
    return build_name_func


@functools.lru_cache(500)
def extract_domain(url: str) -> str:
    """
    Returns the domain
    """
    try:
        import tldextract
        return tldextract.extract(url).registered_domain
    except ImportError:
        from urllib.parse import urlparse
        return urlparse(url).netloc

@functools.lru_cache()
def get_root_domain(url: str, attempts: Optional[int] = None) -> str:
    """
    Returns the root domain after redirection
    """
    if not url.startswith('http'): url = f'https://{url}'
    try:
        with aiohttpx.Client(follow_redirects = True, verify = False) as client:
            r = client.get(url, timeout = 5)
        r.raise_for_status()
        return str(r.url).rstrip('/')
    except (aiohttpx.ConnectTimeout, aiohttpx.ReadTimeout) as e:
        return url
    except aiohttpx.HTTPStatusError as e:
        return str(e.response.url).rstrip('/') if e.response.status_code < 500 else url
    except Exception as e:
        attempts = attempts + 1 if attempts else 1
        if attempts > 2:
            return None
        new_url = f'https://{extract_domain(url)}'
        return get_root_domain(new_url, attempts = attempts)


@alru_cache(maxsize=1200)
async def aget_root_domain(url: str, attempts: Optional[int] = None) -> str:
    """
    Returns the root domain after redirection
    """
    if not url.startswith('http'): url = f'https://{url}'
    try:
        async with aiohttpx.Client(follow_redirects = True, verify = False) as client:
            r = await client.async_get(url, timeout = 5)
        r.raise_for_status()
        return str(r.url).rstrip('/')
    except (aiohttpx.ConnectTimeout, aiohttpx.ReadTimeout) as e:
        return url
    except aiohttpx.HTTPStatusError as e:
        return str(e.response.url).rstrip('/') if e.response.status_code < 500 else url
    except Exception as e:
        attempts = attempts + 1 if attempts else 1
        if attempts > 2:
            return None
        new_url = f'https://{extract_domain(url)}'
        return await aget_root_domain(new_url, attempts = attempts)


def fatal_http_exception(exc):
    """
    Checks if the exception is fatal
    """
    if isinstance(exc, aiohttpx.HTTPStatusError):
        # retry on server errors and client errors
        # with 429 status code (rate limited),
        # don't retry on other client errors
        return (400 <= exc.response.status_code < 500) and exc.response.status_code != 429
    else:
        # retry on all other errors (eg. network)
        return False

http_retry_wrapper = functools.partial(
    backoff.on_exception,
    backoff.expo, 
    exception = Exception, 
    giveup = fatal_http_exception,
    factor = 5,
)





DefaultUserAgents = [
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36',
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36 Edg/116.0.1938.81",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.116 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36 Edg/119.0.0.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.3",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.3",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.3",
]

def get_user_agent(*args, **kwargs) -> str:
    """
    Returns a user agent
    """
    try:
        from fake_useragent import UserAgent
        ua = UserAgent(*args, **kwargs)
        return ua.random
    except ImportError:
        return random.choice(DefaultUserAgents)

    



from __future__ import annotations

"""
API Client that utilizes the AuthZero API
"""

import socket
import contextlib
from abc import ABC
from urllib.parse import urljoin
from lazyops.libs import lazyload
from lazyops.utils.helpers import timed_cache
from lazyops.libs.abcs.configs.types import AppEnv

from ..utils.lazy import get_az_settings, logger, get_az_flow, get_az_resource
from typing import Optional, List, Dict, Any, Union

if lazyload.TYPE_CHECKING:
    import niquests
    import tldextract
    from niquests import Session, AsyncSession, Response, AsyncResponse
    from ..configs import AuthZeroSettings
    from ..flows.tokens import APIClientCredentialsFlow
    from ..flows.admin import AZManagementAPI
    from ..types.auth import AuthZeroTokenAuth
else:
    niquests = lazyload.LazyLoad("niquests")
    tldextract = lazyload.LazyLoad("tldextract")
    from ..types.tokens import AccessToken


def resolve_endpoint(
    endpoints: List[str],
) -> str:
    """
    Resolves the endpoint
    """
    for endpoint in endpoints:
        ep = tldextract.extract(endpoint)
        with contextlib.suppress(Exception):
            if ep.registered_domain:
                socket.gethostbyname(ep.registered_domain)
            else:
                socket.gethostbyaddr(ep.ipv4)
            return endpoint
    raise ValueError(f"Could not resolve endpoint: {endpoints}")


class AuthZeroAPIClient(ABC):

    """
    Generic Auth0 API Client that can be used to access other APIs that require Auth0 Authentication
    """

    name: Optional[str] = None
    api_key: Optional[str] = None
    api_client_id: Optional[str] = None
    api_client_env: Optional[Union[str, AppEnv]] = None

    oauth_url: Optional[str] = None
    client_id: Optional[str] = None
    client_secret: Optional[str] = None
    endpoint: Optional[str] = None
    audience: Optional[str] = None
    session_enabled: Optional[bool] = True
    default_timeout: Optional[float] = 600.0
    default_retries: Optional[int] = 3

    def __init__(
        self, 
        endpoint: Optional[Union[str, List[str]]] = None, 
        api_key: Optional[str] = None,
        api_client_id: Optional[str] = None, 
        api_client_env: Optional[str] = None, 

        audience: Optional[str] = None,
        oauth_url: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        session_enabled: Optional[bool] = None,
        verbose: Optional[bool] = False,
        **kwargs,
    ):
        if endpoint is not None: 
            if isinstance(endpoint, list):
                endpoint = resolve_endpoint(endpoint)
            self.endpoint = endpoint
        if api_key is not None: self.api_key = api_key
        if api_client_id is not None: self.api_client_id = api_client_id
        if api_client_env is not None: self.api_client_env = api_client_env
        
        if oauth_url is not None: self.oauth_url = oauth_url
        if client_id is not None: self.client_id = client_id
        if client_secret is not None: self.client_secret = client_secret
        
        if audience is not None: self.audience = audience
        if session_enabled is not None: self.session_enabled = session_enabled
        assert self.audience is not None, "Audience must be set"
        self.verbose = verbose

        
        self._kwargs = kwargs
        self._settings: Optional['AuthZeroSettings'] = None
        self._token_flow: Optional['APIClientCredentialsFlow'] = None
        self._auth: Optional['AuthZeroTokenAuth'] = None
        self._identity: Optional[str] = None

        self._session: Optional['Session'] = None
        self._asession: Optional['AsyncSession'] = None

        self._data: Optional[Dict[str, Any]] = None
        self._authorized: Optional[bool] = None

    """
    Lazily Initialized Properties
    """
        
    @property
    def settings(self) -> 'AuthZeroSettings':
        """
        Returns the settings
        """
        if self._settings is None: self._settings = get_az_settings()
        return self._settings
    
    @property
    def token_flow(self) -> 'APIClientCredentialsFlow':
        """
        Token Flow
        """
        if self._token_flow is None:
            self._token_flow = get_az_flow(
                name = 'api_client_credentials',
                endpoint = self.endpoint,
                api_client_id = self.api_client_id,
                api_client_env = self.api_client_env,
                audience = self.audience,
                client_id = self.client_id,
                client_secret = self.client_secret,
                oauth_url = self.oauth_url,
            )
        return self._token_flow
    
    @property
    def auth(self) -> 'AuthZeroTokenAuth':
        """
        Auth Object
        """
        if self._auth is None:
            self._auth = get_az_resource(
                'auth_zero_token_auth',
                token_flow = self.token_flow,
            )
            self._auth.headers['accept'] = 'application/json'
            self._auth.headers['content-type'] = 'application/json'
            if self.api_key:
                self._auth.x_api_key = self.api_key
            if self.api_client_id:
                self._auth.headers[self.settings.api_client_id_header_key] = self.api_client_id
            if self.api_client_env:
                self._auth.headers[self.settings.api_client_env_header_key] = self.api_client_env
            if self._kwargs.get('headers'):
                self._auth.headers.update(self._kwargs.get('headers'))
        return self._auth
    
    @property
    def data(self) -> Dict[str, Any]:
        """
        Returns the data
        """
        if self._data is None: self._data = self.token_flow.load_data()
        return self._data


    def get_url(self, path: str) -> str:
        """
        Returns the URL
        """
        return urljoin(self.endpoint, path)
    
    def get_session_kwargs(self, is_async: Optional[bool] = None, **kwargs) -> Dict[str, Any]:
        """
        Returns the session kwargs
        """
        # resolvers: List[str] = kwargs.pop('resolver', [])
        # if not resolvers:
        #     resolvers = []
        #     if self.settings.in_k8s and 'cluster.local' in self.endpoint:
        #         from lazyops.utils.system import fetch_resolver_nameserver
        #         if resolver := fetch_resolver_nameserver():
        #             resolvers.append(f'dou://{resolver}')
        #     resolvers.extend(
        #         [
        #             'doh+cloudflare://',
        #             'doh+google://',
        #         ]
        #     )
        # if resolvers and is_async:
        #     resolvers = resolvers[0]
        # resolvers = resolvers or None
        # logger.info(f'Resolvers: {resolvers}')
        return {
            'pool_connections': self._kwargs.get('pool_connections', 10),
            'pool_maxsize': self._kwargs.get('pool_maxsize', 10),
            'retries': self._kwargs.get('retries', self.default_retries),
            # 'resolver': resolvers,
            **kwargs,
        }

    @property
    def session(self) -> 'Session':
        """
        Returns the Session
        """
        if self._session is None:
            self._session = niquests.Session(**self.get_session_kwargs())
        return self._session
    
    @property
    def asession(self) -> 'AsyncSession':
        """
        Returns the Async Session
        """
        if self._asession is None:
            self._asession = niquests.AsyncSession(**self.get_session_kwargs(is_async=True))
        return self._asession
    
    def reset_session(self):
        """
        Resets the session
        """
        if self._session is not None: self._session.close()
        self._session = None
    
    async def areset_session(self):
        """
        Resets the async session
        """
        if self._asession is not None: await self._asession.close()
        self._asession = None
        if self._session is not None:  self._session.close()
        self._session = None
    
    @timed_cache(600, cache_if_result = True)
    def check_available(self) -> bool:
        """
        Checks if the API is available
        """
        response = None
        try:
            response = niquests.get(self.get_url('/healthz'), timeout = 1.5, retries = 3)
            response.raise_for_status()
            return True
        except Exception as e:
            if self.verbose: logger.warning(f'[{response.status_code} - {self.name}] Error checking API availability: {e}')
            return False
    
        
    @property
    def available(self) -> bool:
        """
        Returns True if the API is available
        """
        return self.check_available()
    
    @property
    def identity(self) -> str:
        """
        Returns the Identity
        """
        if self._identity is None:
            try:
                resp = niquests.get(self.get_url('/healthz'), timeout = 0.5, retries = 3)
                resp.raise_for_status()
                self._identity = resp.text
            except Exception as e:
                if self.verbose: logger.error(f'Error Checking Identity: {e}')
                self._identity = None
        return self._identity


    def check_authorization(self):
        """
        Checks authorization
        """
        if self._authorized: return
        try:
            response = self.session.get(self.get_url('/authorize'), timeout = 2.5, auth = self.auth)
            response.raise_for_status()
        except niquests.HTTPError as e:
            logger.error(f'Error Authorizing Client: {e}')
            return
        
        data: Dict[str, Any] = response.json()
        if api_key := data.get('api_key', data.get('api-key')):
            self._api_key = api_key
            self.data['api_key'] = api_key
            self.auth.x_api_key = api_key
        if api_env := data.get('environment'): self.data['environment'] = api_env
        if identity := data.get('identity'): self.data['identity'] = identity
        self._authorized = True
        self.token_flow.save_data(self.data)

    async def acheck_authorization(self):
        """
        Checks authorization
        """
        if self._authorized: return
        try:
            response = await self.asession.get(self.get_url('/authorize'), timeout = 2.5, auth = self.auth)
            response.raise_for_status()
        except niquests.HTTPError as e:
            logger.error(f'Error Authorizing Client: {e}')
            return
        
        data: Dict[str, Any] = response.json()
        if api_key := data.get('api_key', data.get('api-key')):
            self._api_key = api_key
            self.data['api_key'] = api_key
            self.auth.x_api_key = api_key
        if api_env := data.get('environment'): self.data['environment'] = api_env
        if identity := data.get('identity'): self.data['identity'] = identity
        self._authorized = True
        await self.token_flow.asave_data(self.data)
    
    def preflight_check(self):
        """
        Conducts a pre-flight check
        """
        if not self.available:
            raise ValueError(f"API is not available: {self.endpoint}")
        self.check_authorization()
    
    async def apreflight_check(self):
        """
        Conducts a pre-flight check
        """
        if not self.available:
            raise ValueError(f"API is not available: {self.endpoint}")
        await self.acheck_authorization()
    
    def get_request_kwargs(self, timeout: Optional[float] = None, **kwargs) -> Dict[str, Any]:
        """
        Returns the request kwargs
        """
        auth = kwargs.get('auth', self.auth)
        return {
            'timeout': timeout if timeout is not None else self.default_timeout,
            'auth': auth,
            **kwargs,
        }

    async def aget(self, url: str, *args, timeout: Optional[float] = None, **kwargs) -> 'AsyncResponse':
        """
        Async GET
        """
        await self.apreflight_check()
        kwargs = self.get_request_kwargs(timeout = timeout, **kwargs)
        return await self.asession.get(self.get_url(url), *args, **kwargs)

    async def aput(self, url: str, *args, timeout: Optional[float] = None, **kwargs) -> 'AsyncResponse':
        """
        Async PUT
        """
        await self.apreflight_check()
        kwargs = self.get_request_kwargs(timeout = timeout, **kwargs)
        return await self.asession.put(self.get_url(url), *args, **kwargs)
    
    async def apost(self, url: str, *args, timeout: Optional[float] = None, **kwargs) -> 'AsyncResponse':
        """
        Async POST
        """
        await self.apreflight_check()
        kwargs = self.get_request_kwargs(timeout = timeout, **kwargs)
        return await self.asession.post(self.get_url(url), *args, **kwargs)
    
    async def adelete(self, url: str, *args, timeout: Optional[float] = None, **kwargs) -> 'AsyncResponse':
        """
        Async DELETE
        """
        await self.apreflight_check()
        kwargs = self.get_request_kwargs(timeout = timeout, **kwargs)
        return await self.asession.delete(self.get_url(url), *args, **kwargs)
    
    async def apatch(self, url: str, *args, timeout: Optional[float] = None, **kwargs) -> 'AsyncResponse':
        """
        Async PATCH
        """
        await self.apreflight_check()
        kwargs = self.get_request_kwargs(timeout = timeout, **kwargs)
        return await self.asession.patch(self.get_url(url), *args, **kwargs)
    
    def get(self, url: str, *args, timeout: Optional[float] = None,  **kwargs) -> 'Response':
        """
        GET
        """
        self.preflight_check()
        kwargs = self.get_request_kwargs(timeout = timeout, **kwargs)
        return self.session.get(self.get_url(url), *args, **kwargs)
    
    def put(self, url: str, *args, timeout: Optional[float] = None,  **kwargs) -> 'Response':
        """
        PUT
        """
        self.preflight_check()
        kwargs = self.get_request_kwargs(timeout = timeout, **kwargs)
        return self.session.put(self.get_url(url), *args, **kwargs)
    
    def post(self, url: str, *args, timeout: Optional[float] = None, **kwargs) -> 'Response':
        """
        POST
        """
        self.preflight_check()
        kwargs = self.get_request_kwargs(timeout = timeout, **kwargs)
        return self.session.post(self.get_url(url), *args, **kwargs)
    
    def delete(self, url: str, *args, timeout: Optional[float] = None, **kwargs) -> 'Response':
        """
        DELETE
        """
        self.preflight_check()
        kwargs = self.get_request_kwargs(timeout = timeout, **kwargs)
        return self.session.delete(self.get_url(url), *args, **kwargs)
    
    def patch(self, url: str, *args, timeout: Optional[float] = None, **kwargs) -> 'Response':
        """
        PATCH
        """
        self.preflight_check()
        kwargs = self.get_request_kwargs(timeout = timeout, **kwargs)
        return self.session.patch(self.get_url(url), *args, **kwargs)
    


from __future__ import annotations

"""
This module contains a wrapper function that allows you to wrap an object class to create a proxy object.


@proxied
class DummyClass(abc.ABC):

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        print('DummyClass initialized')

    def hi(self, *args, **kwargs):
        print('DummyClass called')
        return 'hi'
        

# Testing a proxy object wrapper with arguments
x = DummyClass
>> DummyClass initialized

print(x.hi())
>> DummyClass called
>> hi
"""

from typing import Type, TypeVar, Optional, Union, Any, Callable, List, Dict, overload, TYPE_CHECKING
from .main import ProxyObject

ObjT = TypeVar('ObjT')
ProxyObjT = Type[ObjT]


@overload
def proxied(
    obj_cls: Optional[ProxyObjT] = None,
    obj_getter: Optional[Union[Callable, str]] = None,
    obj_args: Optional[List[Any]] = None,
    obj_kwargs: Optional[Dict[str, Any]] = None,
    obj_initialize: Optional[bool] = True,
    threadsafe: Optional[bool] = True,
) -> Union[ObjT, ProxyObjT]: 
    """
    Lazily initialize an object as a proxy object

    Args:
    - obj_cls: The object class
    - obj_getter: The object getter. This can be a callable or a string which will lazily import the object
    - obj_args: The object arguments. This can be a list or a callable that returns a list, or a string that lazily imports a function/list
    - obj_kwargs: The object keyword arguments. This can be a dictionary or a callable that returns a dictionary, or a string that lazily imports a function/dictionary
    - obj_initialize: The object initialization flag
    - threadsafe: The thread safety flag

    Returns:
    - Type[OT]: The proxy object

    Usage:
    @proxied
    class DummyClass(abc.ABC):

        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs
            print('DummyClass initialized')
    
    x = DummyClass
    >> DummyClass initialized

    # Note that if you initialize the object directly, it will raise an error
    y = DummyClass() # Raises an error

    def dummy_args_getter() -> Iterable:
        return [1, 2]

    # @proxied(obj_args=(1, 2), obj_kwargs={'a': 1, 'b': 2})
    @proxied(obj_args=dummy_args_getter, obj_kwargs={'a': 1, 'b': 2})
    class DummyClass(abc.ABC):

        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs
            print('DummyClass initialized', args, kwargs)
    
    x = DummyClass
    >> DummyClass initialized (1, 2) {'a': 1, 'b': 2}
    """
    ...


@overload
def proxied(
    **kwargs: Any,
) -> Callable[..., Union[ObjT, ProxyObjT]]:
    """
    Lazily initialize an object as a proxy object

    Args:
    - obj_cls: The object class
    - obj_getter: The object getter. This can be a callable or a string which will lazily import the object
    - obj_args: The object arguments. This can be a list or a callable that returns a list, or a string that lazily imports a function/list
    - obj_kwargs: The object keyword arguments. This can be a dictionary or a callable that returns a dictionary, or a string that lazily imports a function/dictionary
    - obj_initialize: The object initialization flag
    - threadsafe: The thread safety flag

    Returns:
    - Type[OT]: The proxy object

    Usage:
    @proxied
    class DummyClass(abc.ABC):

        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs
            print('DummyClass initialized')
    
    x = DummyClass
    >> DummyClass initialized

    # Note that if you initialize the object directly, it will raise an error
    y = DummyClass() # Raises an error

    def dummy_args_getter() -> Iterable:
        return [1, 2]

    # @proxied(obj_args=(1, 2), obj_kwargs={'a': 1, 'b': 2})
    @proxied(obj_args=dummy_args_getter, obj_kwargs={'a': 1, 'b': 2})
    class DummyClass(abc.ABC):

        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs
            print('DummyClass initialized', args, kwargs)
    
    x = DummyClass
    >> DummyClass initialized (1, 2) {'a': 1, 'b': 2}
    """
    def wrapper(obj_cls: ProxyObjT) -> Union[ObjT, ProxyObjT]:
        ...
    return wrapper



def proxied(
    obj_cls: Optional[ProxyObjT] = None,
    obj_getter: Optional[Union[Callable, str]] = None,
    obj_args: Optional[List[Any]] = None,
    obj_kwargs: Optional[Dict[str, Any]] = None,
    obj_initialize: Optional[bool] = True,
    threadsafe: Optional[bool] = True,
) -> Union[Callable[..., Union[ObjT, ProxyObjT]], Union[ObjT, ProxyObjT]]:
    """
    Lazily initialize an object as a proxy object

    Args:
    - obj_cls: The object class
    - obj_getter: The object getter. This can be a callable or a string which will lazily import the object
    - obj_args: The object arguments. This can be a list or a callable that returns a list, or a string that lazily imports a function/list
    - obj_kwargs: The object keyword arguments. This can be a dictionary or a callable that returns a dictionary, or a string that lazily imports a function/dictionary
    - obj_initialize: The object initialization flag
    - threadsafe: The thread safety flag

    Returns:
    - Type[OT]: The proxy object

    Usage:
    @proxied
    class DummyClass(abc.ABC):

        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs
            print('DummyClass initialized')
    
    x = DummyClass
    >> DummyClass initialized

    # Note that if you initialize the object directly, it will raise an error
    y = DummyClass() # Raises an error

    def dummy_args_getter() -> Iterable:
        return [1, 2]

    # @proxied(obj_args=(1, 2), obj_kwargs={'a': 1, 'b': 2})
    @proxied(obj_args=dummy_args_getter, obj_kwargs={'a': 1, 'b': 2})
    class DummyClass(abc.ABC):

        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs
            print('DummyClass initialized', args, kwargs)
    
    x = DummyClass
    >> DummyClass initialized (1, 2) {'a': 1, 'b': 2}
    """

    if obj_cls is not None:
        return ProxyObject(
            obj_cls = obj_cls,
            obj_getter = obj_getter,
            obj_args = obj_args,
            obj_kwargs = obj_kwargs,
            obj_initialize = obj_initialize,
            threadsafe = threadsafe,
        )
    
    def wrapper(obj_cls: ProxyObjT) -> Union[ObjT, ProxyObjT]:
        return ProxyObject(
            obj_cls = obj_cls,
            obj_getter = obj_getter,
            obj_args = obj_args,
            obj_kwargs = obj_kwargs,
            obj_initialize = obj_initialize,
            threadsafe = threadsafe,
        )
    
    return wrapper
